//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this sample source code is subject to the terms of the Microsoft
// license agreement under which you licensed this sample source code. If
// you did not accept the terms of the license agreement, you are not
// authorized to use this sample source code. For the terms of the license,
// please see the license agreement between you and Microsoft or, if applicable,
// see the LICENSE.RTF on your install media or the root of your tools installation.
// THE SAMPLE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//

#include <windows.h>
#include <guts.h>
#include <ehm.h>
#include <tchddi.h>
#include <sync.hxx>
#include <touchperf.h>
#include <mdpgperf.h>
#include <regkey.h>
#include <prsht.h>
#include "resource.h"
#include "shellapi.h"

#include "cetouchview.h"
#include <list>

// CETouchView requires that the touch driver is a new-style driver.
//
// CETouchFilter must be inserted into the touch stack between the touch driver and tchproxy. It will place
// the touch points into a message queue for CETouchView as well as passing them on to tchproxy.
//
// To deploy:
// 1. To load the CETouchFilter driver, add/change the following registry entries:
//    [HKEY_LOCAL_MACHINE\HARDWARE\DEVICEMAP\TOUCH]
//        "DriverName"="CETouchFilter.dll"
//        "DriverExName"="tchproxy.dll"
//
// 2. Reboot
//
// This will replace the tchproxy with CETouchFilter. CETouchFilter will then load tchproxy.
//
// NOTE: if your configuration does not use tchproxy, replace DriverExName with the appropriate DLL.
//

// Uncomment this line for increased debug output
//#define DEBUG_OUTPUT 1
//

#define TOUCHPOINT_RADIUS       (5)             // radius of the ellipse showing the touch point

// Performance related
//
LPCTSTR PERFSCENARIO_DLL = L"PerfScenario.dll";
GUID touchGUID = { 0x25121442, 0x8ca, 0x48dd, { 0x91, 0xcc, 0xbf, 0xc9, 0xf7, 0x90, 0x2, 0x7c } };
#define PERF_GESTURE_NAMESPACE L"Gesture"
typedef HRESULT (*PFN_PERFSCEN_OPENSESSION)(LPCTSTR lpszSessionName, BOOL fStartRecordingNow);
typedef HRESULT (*PFN_PERFSCEN_CLOSESESSION)(LPCTSTR lpszSessionName);
typedef HRESULT (*PFN_PERFSCEN_FLUSHMETRICS)(BOOL fCloseAllSessions, GUID* scenarioGuid, LPCTSTR lpszScenarioNamespace, LPCTSTR lpszScenarioName, LPCTSTR lpszLogFileName, LPCTSTR lpszTTrackerFileName, GUID* instanceGuid);

HINSTANCE g_hLibPerfScenario = NULL;
LPVOID pCePerfInternal;        // Instantiate REQUIRED CePerf global
PFN_PERFSCEN_OPENSESSION  g_lpfnPerfScenarioOpenSession  = NULL;
PFN_PERFSCEN_CLOSESESSION g_lpfnPerfScenarioCloseSession = NULL;
PFN_PERFSCEN_FLUSHMETRICS g_lpfnPerfScenarioFlushMetrics = NULL;

#if !defined(LODWORD)
typedef unsigned __int64 QWORD;
#define LODWORD(qw) (DWORD)(0xffffffff&(qw))
#define HIDWORD(qw) (DWORD)(((QWORD)(qw))>>32)
#define MAKEQWORD(dwLow,dwHigh) (((QWORD)(dwLow))|(((QWORD)(dwHigh))<<32))
#endif

struct FLAG_TO_STRING
{
    DWORD dwFlag;
    LPCTSTR psz;
};

typedef unsigned __int64 QWORD;
#define HIDWORD(qw) (DWORD)(((QWORD)(qw))>>32)

// As new gestures are added, all that needs to be updated is this table.
static const FLAG_TO_STRING s_rgGestureToString[] = {
    {GID_BEGIN,                 TEXT("BEGIN") },
    {GID_END,                   TEXT("END") },
    {GID_PAN,                   TEXT("PAN") },
    {GID_ROTATE,                TEXT("ROTATE") },
    {GID_SCROLL,                TEXT("SCROLL") },
    {GID_HOLD,                  TEXT("HOLD") },
    {GID_SELECT,                TEXT("SELECT") },
    {GID_DOUBLESELECT,          TEXT("DOUBLE SELECT") },
    {GID_DIRECTMANIPULATION,    TEXT("DIRECT MANIPULATION") }
};


static const FLAG_TO_STRING s_rgFlagToString[] = {
    { TOUCHEVENTF_MOVE,         TEXT("MV") },       // MOVE
    { TOUCHEVENTF_DOWN,         TEXT("DN") },       // DOWN
    { TOUCHEVENTF_UP,           TEXT("UP") },       // UP
    { TOUCHEVENTF_INRANGE,      TEXT("IR") },       // INRANGE
    { TOUCHEVENTF_PRIMARY,      TEXT("PR") },       // PRIMARY
    { TOUCHEVENTF_NOCOALESCE,   TEXT("NC") },       // NOCOALESCE
    { TOUCHEVENTF_PEN,          TEXT("PN") },       // PEN
    { TOUCHEVENTF_PALM,         TEXT("PM") },       // PALM
    { TOUCHEVENTF_CALIBRATED,   TEXT("CL") },       // CALIBRATED
    { TOUCHEVENTF_SYMMETRIC,    TEXT("SM") },       // SYMMETRIC
};

const TCHAR* const szWndClassName = TEXT("CETouchViewClass");
const TCHAR* const szMainWndTitle = TEXT("CETouchView");

static volatile CETOUCHINPUT s_rgInputs[CETOUCHINPUT_MAX_SIMULTANEOUS];
static volatile size_t s_cInputs;

typedef struct
{
    size_t cInputs;
    CETOUCHINPUT rgInputs[CETOUCHINPUT_MAX_SIMULTANEOUS];
} TouchInputData;

std::list<TouchInputData *> RawList;

static ce::critical_section s_cs;

static HANDLE s_hMsgQ;
HWND g_hwnd = NULL;
HDC g_hdcPerm = NULL;
HWND g_hwndTaskbar = NULL;
int g_origTaskbarY = 0;

int g_dwScreeW;
int g_dwScreeH;

static HANDLE s_hExit;

static GESTUREINFO CurrentGesture = {0};
static GESTUREINFO LastBeginGesture = {0};

// Zooming
static DWORD s_dwZoomLevel = 10;

#define MAX_ZOOM_LEVEL  24
#define MIN_ZOOM_LEVEL  1
#define MAX_FONT_SIZE   180
#define MIN_FONT_SIZE   12

// GDI objects

static HFONT s_hBoldFont = NULL, s_hDemoFonts[MAX_ZOOM_LEVEL];

// Using PS (Pan/Scroll) or DM
static BOOL s_fUseDM = FALSE;

// Driver sent an Up event
static BOOL s_fUpWarning = FALSE;

// Driver sent duplicate dwIDs
static BOOL s_fIDWarning = FALSE;

static DWORD s_dwDrawModeIdx = 0;

// Gesture count
DWORD g_dwMsgCount_Gesture = 0;
DWORD g_dwGesturesInSession = 0;
DWORD g_dwTimeInSession = 0;
DWORD g_dwMsgCount_Gesture_On_Begin = 0;
DWORD g_dwTime_On_Begin;
DWORD g_dwGesturesPerSecond = 0;
DWORD g_dwGestureTypeCount[13];

REGSETTINGS g_Settings;
static RECT s_rcClient = {0};

static volatile BOOL s_fResetRegistry = FALSE;       // debug aid, set to TRUE to start reset values - can be done in debugger

HBRUSH g_rgBrushes[NUM_BRUSHES];

void PaintCtrlArea(HDC hdc);
BOOL SwapDM (HWND hwnd, BOOL fUseDM);
BOOL InitScreenObjects();
void DrawGestureDot ( GESTUREINFO *pgi );
void DrawGestureText ( GESTUREINFO *pgi );
void DrawGestureBar ( GESTUREINFO *pgi );
void DrawRawDots(void);
void PanProc(GESTUREINFO *pgi);
void ShowStatistics ( void );

HINSTANCE g_hInstance = NULL;
LPCTSTR LoadStringZ ( UINT id );
BOOL g_fFilterPresent = FALSE;
static HANDLE s_hEvtAutoData = NULL;
BOOL g_fRecording = FALSE;
HWND g_hwndButton = NULL;
HWND g_hwndData = NULL;
HBITMAP g_hBkImage = NULL;
BITMAP g_bmBkImage = {0};

void WarningMessage ( LPCTSTR szMessage, BOOL fBeep = TRUE );

void WarningMessage ( LPCTSTR szMessage, BOOL fBeep )
{
    //RETAILMSG(TRUE,(szMessage));

    if (fBeep)
    {
        MessageBeep(MB_ICONEXCLAMATION);
    }

    if (0 == g_Settings.dwShowData)
    {
        return;
    }

    SetWindowText(g_hwndData, szMessage);
}

LPCTSTR FlagToString ( FLAG_TO_STRING *rgFlags, size_t rgCount, DWORD dwValue )
{
    for (size_t idx = 0; idx < rgCount; ++idx)
    {
        if (rgFlags[idx].dwFlag == dwValue)
        {
            return rgFlags[idx].psz;
        }
    }

    return NULL;
}

LPTSTR FlagsToString ( FLAG_TO_STRING *rgFlags, size_t rgCount, DWORD dwFlags )
{
    WCHAR szTmp[256];
    DWORD dwBit = 1;
    LPCTSTR szFlag = NULL;

    *szTmp = L'\0';
    while (dwFlags)
    {
        if (dwFlags & dwBit)
        {
            szFlag = FlagToString(rgFlags, rgCount, dwBit);
            if (NULL == szFlag)
            {
                WCHAR szUnk[16];
                StringCchPrintf(szUnk, _countof(szUnk), L"0x%08X", dwBit);
                StringCchCat(szTmp, _countof(szTmp), szUnk);
            }
            else
            {
                StringCchCat(szTmp, _countof(szTmp), szFlag);
            }

            dwFlags -= dwBit;
            if (dwFlags)
            {
                StringCchCat(szTmp, _countof(szTmp), L" | ");
            }
        }

        dwBit <<= 1;
    }

    return StrDup(szTmp);
}

// Call this to update the data window.
// designed to be called with either or both gesture and raw data
// if you pass null update will use previously passed data
// this keeps a local copy of data
void UpdateDataWindow ( GESTUREINFO *pgi, TouchInputData *ptd )
{
    static GESTUREINFO mygi = {0};
    static TouchInputData mytd = {0};

    // If we're turned off just return without doing anything...
    if (0 == g_Settings.dwShowData)
    {
        return;
    }

    // If both are NULL, erase all data
    if ((NULL == pgi) && (NULL == ptd))
    {
        ZeroMemory(&mygi, sizeof(GESTUREINFO));
        ZeroMemory(&mytd, sizeof(TouchInputData));
        SetWindowText(g_hwndData, L"");
        return;
    }

    // Only update gesture data if it's given
    if (NULL != pgi)
    {
        memcpy(&mygi, pgi, sizeof(GESTUREINFO));
    }

    // Only update raw data if it's given
    if (NULL != ptd)
    {
        memcpy(&mytd, ptd, sizeof(ptd->cInputs) + ((min(ptd->cInputs,CETOUCHINPUT_MAX_SIMULTANEOUS) * sizeof(*ptd->rgInputs))));
    }

    WCHAR szTm1[1024], szTm2[256];

    if (0 == mygi.cbSize)
    {
        *szTm1 = L'\0';
    }
    else
    {
        StringCchPrintf(szTm1, _countof(szTm1),
            L"Gesture: %-20s  X,Y: %4d,%4d  Inst: %08X  Seq: %08X Args: %016I64X",
            FlagToString((FLAG_TO_STRING *)s_rgGestureToString, _countof(s_rgGestureToString), mygi.dwID),
            mygi.ptsLocation.x,
            mygi.ptsLocation.y,
            mygi.dwInstanceID,
            mygi.dwSequenceID,
            mygi.ullArguments);
    }

    StringCchCat(szTm1, _countof(szTm1), L"\n");

    *szTm2 = L'\0';
    for (size_t i = 0; i < mytd.cInputs; ++i)
    {
        LPTSTR szTmp = NULL;    // Keep pointer around to be able to free the allocated string

        StringCchPrintf(szTm2, _countof(szTm2),
            L"Raw[%d]: X,Y: %4d,%4d  Src:%08X ID:%08X Mask:%08X Time:%08X CX,CY:%3d,%3d Flags: %s",
            i+1,
            mytd.rgInputs[i].x,
            mytd.rgInputs[i].y,
            mytd.rgInputs[i].hSource,
            mytd.rgInputs[i].dwID,
            mytd.rgInputs[i].dwMask,
            mytd.rgInputs[i].dwTime,
            mytd.rgInputs[i].cxContact,
            mytd.rgInputs[i].cyContact,
            szTmp = FlagsToString((FLAG_TO_STRING *)s_rgFlagToString, _countof(s_rgFlagToString), mytd.rgInputs[i].dwFlags));

        LocalFree(szTmp);
        StringCchCat(szTm1, _countof(szTm1), szTm2);
        StringCchCat(szTm1, _countof(szTm1), L"\n");
    }

    SetWindowText(g_hwndData, szTm1);
}

LPWSTR StrDup ( LPWSTR lpstr )
{
    LPWSTR p = NULL;
    size_t len = 0;
    if (SUCCEEDED(StringCchLength(lpstr, STRSAFE_MAX_CCH, &len)))
    {
        p = (LPWSTR)LocalAlloc(LPTR, (len + 1) * sizeof(WCHAR));
        if (NULL != p)
        {
            StringCchCopy(p, len + 1, lpstr);
        }
    }

    return p;
}

void InitPerf()
{
    g_hLibPerfScenario = LoadLibrary(PERFSCENARIO_DLL);
    if (g_hLibPerfScenario)
    {
        g_lpfnPerfScenarioOpenSession  = reinterpret_cast<PFN_PERFSCEN_OPENSESSION>  (GetProcAddress(g_hLibPerfScenario, L"PerfScenarioOpenSession"));
        g_lpfnPerfScenarioCloseSession = reinterpret_cast<PFN_PERFSCEN_CLOSESESSION> (GetProcAddress(g_hLibPerfScenario, L"PerfScenarioCloseSession"));
        g_lpfnPerfScenarioFlushMetrics = reinterpret_cast<PFN_PERFSCEN_FLUSHMETRICS> (GetProcAddress(g_hLibPerfScenario, L"PerfScenarioFlushMetrics"));
    }
    TouchPerf::AcquireSession();

    if (g_lpfnPerfScenarioOpenSession)
    {
        (*g_lpfnPerfScenarioOpenSession)(MDPGPERF_TOUCH, TRUE);
    }
}


void ExitPerf()
{
    if (g_lpfnPerfScenarioFlushMetrics)
    {
        // Flush CePerf/PerfScenario data
        (*g_lpfnPerfScenarioFlushMetrics)(
                FALSE,
                &touchGUID,
                PERF_GESTURE_NAMESPACE,
                TEXT("GESTURE PERF"),
                L"\\Release\\CETouchView.xml",
                NULL,
                NULL );
    }

    if (g_lpfnPerfScenarioOpenSession)
    {
        (*g_lpfnPerfScenarioCloseSession)(MDPGPERF_TOUCH);
    }

    if (g_hLibPerfScenario)
    {
        FreeLibrary(g_hLibPerfScenario);
    }

    TouchPerf::ReleaseSession();
}

BOOL UpdateBkBitmap ( LPTSTR szFileName )
{
    if (NULL != g_hBkImage)
    {
        DeleteObject(g_hBkImage);
        g_hBkImage = NULL;
        ZeroMemory(&g_bmBkImage, sizeof(g_bmBkImage));
    }

    g_hBkImage = SHLoadDIBitmap ( szFileName );
    if (NULL == g_hBkImage)
    {
        return FALSE;
    }

    GetObject(g_hBkImage, sizeof(g_bmBkImage), &g_bmBkImage);
    return TRUE;
}

void OnTouchGesture ( HGESTUREINFO hgi )
{
    GESTUREINFO gi = { 0 };
    gi.cbSize = sizeof(GESTUREINFO);

    if (!GetGestureInfo(hgi, &gi))
    {
        // Error!  This should never fail.  Invoke debugger for now...
        DebugBreak();
        return;
    }

    ++g_dwMsgCount_Gesture;

    g_dwGestureTypeCount[gi.dwID]++;

    switch (g_Settings.dwDisplayMode)
    {
        case DISPLAYMODE_BOTH:
            DrawRawDots();      // intentionally fall through to gesture
        case DISPLAYMODE_GESTURE:
            DrawGestureDot(&gi);
            break;
        case DISPLAYMODE_TEXT:
            DrawGestureText(&gi);
            break;
        case DISPLAYMODE_BAR:
            DrawGestureBar(&gi);
            break;
        case DISPLAYMODE_RAW:
            DrawRawDots();
            break;
        default:
            // Error: invalid mode, invoke debugger for now...
            DebugBreak();
    }

    UpdateDataWindow(&gi, NULL);
    switch (gi.dwID)
    {
        case GID_BEGIN:
            s_fUpWarning = FALSE;
            memset(&CurrentGesture, 0, sizeof (CurrentGesture));
            memset(g_dwGestureTypeCount, 0, sizeof(g_dwGestureTypeCount));

            g_dwMsgCount_Gesture_On_Begin = g_dwMsgCount_Gesture;
            g_dwTime_On_Begin = GetTickCount();
            LastBeginGesture = gi;
            break;

        case GID_END:
            s_fUpWarning = FALSE;
            g_dwGesturesInSession = g_dwMsgCount_Gesture - g_dwMsgCount_Gesture_On_Begin - 1;
            g_dwTimeInSession = GetTickCount() - g_dwTime_On_Begin;
            g_dwGesturesPerSecond = (g_dwTimeInSession != 0)
                ? (g_dwGesturesInSession * 1000 / g_dwTimeInSession)
                : 99999;

            PostMessage(g_hwnd, WM_SHOW_STATISTICS, 0, 0);

            if (0 == g_Settings.dwShowTrailing)
            {
                ClearScreen();
            }
            break;

        default:
            // copy gesture
            CurrentGesture = gi;
            PanProc(&gi);
            break;
    } // end switch

    // show dwID warning if enabled and not found
    if (g_Settings.dwShowIDWarning  && s_fIDWarning)
    {
//        MessageBox(g_hwnd, L"Duplicate dwID found!", L"dwID Warning enabled", MB_OK);
        WarningMessage(L"\nWarning, Duplicate dwID found!");
        s_fIDWarning = FALSE;
    }

}

void ShowStatistics ( void )
{
    if (g_Settings.dwShowStatistics)
    {
        WCHAR szTmp[256];

        StringCchPrintf(szTmp, _countof(szTmp),
            L"Statistics\n%d G/sec\tCount: %d in %dms\nPan: %d     Scroll: %d     Direct Manipulation: %d     Hold: %d     Select: %d     Double Select: %d",
            g_dwGesturesPerSecond, g_dwGesturesInSession, g_dwTimeInSession,
            g_dwGestureTypeCount[GID_PAN],
            g_dwGestureTypeCount[GID_SCROLL],
            g_dwGestureTypeCount[GID_DIRECTMANIPULATION],
            g_dwGestureTypeCount[GID_HOLD],
            g_dwGestureTypeCount[GID_SELECT],
            g_dwGestureTypeCount[GID_DOUBLESELECT]);
        WarningMessage(szTmp, FALSE);
        ClearRawList();
    }
}

void ClearScreen ( void )
{
    if (BKMODE_IMAGE == g_Settings.dwBkMode)
    {
        ASSERT(g_hBkImage);

        HDC hdcTemp = CreateCompatibleDC(NULL);
        HBITMAP hbmpOld = (HBITMAP)SelectObject(hdcTemp, g_hBkImage);

        StretchBlt( g_hdcPerm,
                    0, 0,
                    g_dwScreeW, g_dwScreeH,
                    hdcTemp,
                    0, 0,
                    g_bmBkImage.bmWidth, g_bmBkImage.bmHeight,
                    SRCCOPY);
        SelectObject(hdcTemp, hbmpOld);
        DeleteDC(hdcTemp);
    }
    else if (BKMODE_COLOR == g_Settings.dwBkMode)
    {
        RECT rc = {0, 0, g_dwScreeW, g_dwScreeH};
        HBRUSH hbr = CreateSolidBrush(g_Settings.dwBkColor);

        FillRect(g_hdcPerm, &rc, hbr);
        DeleteObject(hbr);
    }

    InvalidateRect(g_hwnd, NULL, TRUE);

    UpdateDataWindow(NULL, NULL);
}
int OnPaint ( HWND hWnd, WPARAM wParam, LPARAM lParam )
{
    PAINTSTRUCT ps;

    BeginPaint(hWnd, &ps);

    BitBlt( ps.hdc,
            0, 0,
            g_dwScreeW, g_dwScreeH,
            g_hdcPerm,
            0, 0,
            SRCCOPY);

    EndPaint(hWnd, &ps);
    return 0;
}

LRESULT CALLBACK
MainWndProc(
            HWND    hwnd,
            UINT    message,
            WPARAM  wParam,
            LPARAM  lParam
            )
{
    HBITMAP hbmMem;
    HDC hDC;

    switch ( message )
    {
        case WM_SHOW_CONTACTSWARNING:
            if (g_Settings.dwShowContactsWarning)
            {
                POINT pt;
                POINTSTOPOINT(pt,LastBeginGesture.ptsLocation);
                RECT rc;
                GetWindowRect(g_hwndButton, &rc);
                if (!PtInRect(&rc,pt))
                {
                    if (0 != wParam)
                    {
                        WarningMessage(L"\nContacts Warning: Driver is not sending contact width and height!");
                    }
                    else
                    {
                        WarningMessage(L"\nContacts Warning: Contact information out of declared range!");
                    }
                }
            }
            break;

        case WM_SHOW_STATISTICS:
            if (g_Settings.dwShowStatistics)
            {
                POINT pt;
                POINTSTOPOINT(pt,LastBeginGesture.ptsLocation);
                RECT rc;
                GetWindowRect(g_hwndButton, &rc);
                if (!PtInRect(&rc,pt))
                {
                    ShowStatistics();
                }
            }
            break;

        case WM_TRIGGER_FILTER:
            if (g_fFilterPresent)
            {
                SetEvent(s_hEvtAutoData);
            }
            if (g_fRecording)
            {
                SetWindowText(g_hwndButton, L"Stop Recording");
            }
            break;

        case WM_CREATE:

            hDC = GetDC(NULL);
            g_hdcPerm = CreateCompatibleDC(hDC);
            hbmMem = CreateCompatibleBitmap(hDC, g_dwScreeW, g_dwScreeH);
            SelectObject(g_hdcPerm, hbmMem);
            DeleteObject(hbmMem);
            DeleteDC(hDC);
            ClearScreen();

            if (0 != g_Settings.dwShowDM)
            {
                ASSERT(0 == g_Settings.dwShowPAN);
                ASSERT(0 == g_Settings.dwShowSCROLL);
                SwapDM(hwnd, TRUE);
            }
            else
            {
                SwapDM(hwnd, FALSE);
            }
            break;

        case WM_GESTURE:
            OnTouchGesture(reinterpret_cast<HGESTUREINFO>(lParam));
            break;

        case WM_COMMAND:
            if (BUTTON_ID == LOWORD(wParam))
            {
                if (g_fRecording)
                {
                    SetWindowText(g_hwndButton, L"Options");
                    SetEvent(s_hEvtAutoData);
                    g_fRecording = FALSE;
                }
                else
                {
                    OptionsDialog(hwnd);
                }
            }
            break;

        case WM_PAINT:
            return OnPaint(hwnd,wParam,lParam);
            break;

        case WM_DESTROY:
            if (g_fFilterPresent)
            {
                CloseHandle(s_hEvtAutoData);
            }
            if (NULL != g_hwndTaskbar)
            {
                SetWindowPos(g_hwndTaskbar, NULL, 0, g_origTaskbarY, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
            }
            PostQuitMessage(0);
            return 0;
    }

    return DefWindowProc(hwnd, message, wParam, lParam);
}


BOOL
InitApplication (
                 HINSTANCE hInstance
                 )
{
    WNDCLASS  wc;

    wc.style = 0;
    wc.lpfnWndProc = (WNDPROC)MainWndProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInstance;
    wc.hIcon = 0;
    wc.hCursor = 0;
    wc.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH);
    wc.lpszMenuName = NULL;
    wc.lpszClassName = szWndClassName;

    if (!RegisterClass(&wc))
        return FALSE;

    return InitScreenObjects();
}


BOOL
InitInstance(
             HINSTANCE  hInstance,
             int            nCmdShow,
             HWND       *phwnd
             )
{
    HWND    hwnd;

    g_hwnd = hwnd = CreateWindowEx(
        WS_EX_TOPMOST,
        szWndClassName,
        szMainWndTitle,
        WS_VISIBLE,
        0, 0,
        g_dwScreeW, g_dwScreeH,
        NULL,
        NULL,
        hInstance,
        NULL
        );

    if ( hwnd == 0 )
    {
        return (FALSE);
    }

    // hide task bar by repositioning it off the screen, save original location for when we end
    g_hwndTaskbar = FindWindow(L"HHTaskBar", NULL);
    if (NULL != g_hwndTaskbar)
    {
        RECT rc;
        GetClientRect(g_hwndTaskbar, &rc);
        g_origTaskbarY = g_dwScreeH - rc.bottom + rc.top;
        SetWindowPos(g_hwndTaskbar, NULL, 0, g_dwScreeH + 1, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
    }

    if ( !IsWindow(hwnd) )
    {
        return (FALSE);
    }

    g_hwndButton = CreateWindowEx (
        0,
        L"button",
        L"Options",
        WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
        g_Settings.dwButtonX, g_Settings.dwButtonY,
        BUTTON_WIDTH, BUTTON_HEIGHT,
        hwnd,
        (HMENU)BUTTON_ID,
        hInstance,
        NULL
        );

    g_hwndData = CreateWindowEx (
        0,
        L"static",
        L"",
        WS_CHILD | WS_VISIBLE | WS_BORDER | SS_CENTER,
        g_Settings.dwWindowX, g_Settings.dwWindowY,
        DATAWIN_WIDTH, DATAWIN_HEIGHT,
        hwnd,
        (HMENU)WINDOW_ID,
        hInstance,
        NULL
        );

    ShowWindow(g_hwndData, 0 == g_Settings.dwShowData ? SW_HIDE : SW_SHOW);

    UpdateWindow(hwnd);
    *phwnd = hwnd;

    return(TRUE);
}



static
VOID
SubmitSampleSet(
    DWORD cInputs,
    PCCETOUCHINPUT pInputs
    )
{
    ASSERT(cInputs <= CETOUCHINPUT_MAX_SIMULTANEOUS);

    ce::gate<ce::critical_section> gate(s_cs);

    TouchInputData *td = (TouchInputData *)LocalAlloc(LPTR, sizeof(TouchInputData));
    td->cInputs = cInputs;
    memcpy((LPVOID) td->rgInputs, pInputs, sizeof(*pInputs) * cInputs);

    for (DWORD i = 0; i < cInputs; ++i)
    {
        if (TOUCHEVENTF_UP & pInputs[i].dwFlags)
        {
            s_fUpWarning = TRUE;
        }
    }

    RawList.push_back(td);
}


DWORD WINAPI PointsThread ( LPVOID )
{
    DWORD dwRead;
    DWORD dwFlags;

    CETOUCHINPUT rgInputs[CETOUCHINPUT_MAX_SIMULTANEOUS];
    C_ASSERT(sizeof(rgInputs) == sizeof(s_rgInputs));

    //CeSetThreadPriority(GetCurrentThread(), 109);

    // Empty the queue to avoid replaying old events
    while (ReadMsgQueue(s_hMsgQ, rgInputs, sizeof(rgInputs), &dwRead, 0, &dwFlags))
    {
        ;
    }

    HANDLE rghWaits[] = { s_hMsgQ, s_hExit };

    while ( WAIT_OBJECT_0 == WaitForMultipleObjects(_countof(rghWaits), rghWaits, FALSE, INFINITE) )
    {
        BOOL fSuccess = ReadMsgQueue(s_hMsgQ, rgInputs, sizeof(rgInputs), &dwRead, 0, &dwFlags);

        DWORD cInputs = 0;
        if (fSuccess && (0 == dwRead % sizeof(CETOUCHINPUT)) )
        {
            cInputs = dwRead / sizeof(CETOUCHINPUT);
        }
        // else cause ellipses to be cleared if no contacts present

        SubmitSampleSet(cInputs, rgInputs);
    }

    return 0;
}

void LoadSettings ( void )
{
    CRegKey hKeyTouch;
    ULONG lLen;
    WCHAR szTmp[MAX_PATH];

    // SystemParametersInfo(SPI_GETWORKAREA, 0, &s_rcClient, 0);

    g_dwScreeW = GetSystemMetrics(SM_CXSCREEN);     // s_rcClient.right - s_rcClient.left;
    g_dwScreeH = GetSystemMetrics(SM_CYSCREEN);     // s_rcClient.bottom - s_rcClient.top;

    SetRect(&s_rcClient, 0, 0, g_dwScreeW, g_dwScreeH);

    ZeroMemory(&g_Settings, sizeof(g_Settings));

    if (s_fResetRegistry)
    {
        hKeyTouch.Create(HKEY_LOCAL_MACHINE, L"Software\\Microsoft", NULL, 0, 0, NULL, NULL);
        hKeyTouch.RecurseDeleteKey(L"CETouchView");
        hKeyTouch.Close();
        s_fResetRegistry = FALSE;
    }

    // Create reg key if it doesn't exist, or open it if it does:
    hKeyTouch.Create(HKEY_LOCAL_MACHINE, L"Software\\Microsoft\\CETouchView", NULL, 0, 0, NULL, NULL);

    // Use these nifty macros to make initializing all the reg values easy
#define RegInitDW(x, y) if (ERROR_SUCCESS != hKeyTouch.QueryValue(g_Settings.dw##x, TEXT(#x))) \
                        { \
                            hKeyTouch.SetValue(g_Settings.dw##x = y, TEXT(#x)); \
                        }
#define RegInitSZ(x)    lLen = _countof(szTmp); \
                        if (ERROR_SUCCESS == hKeyTouch.QueryStringValue(TEXT(#x), szTmp, &lLen)) \
                        { \
                            g_Settings.sz##x = StrDup(szTmp); \
                        } \
                        else \
                        { \
                            g_Settings.sz##x = StrDup(L""); \
                            hKeyTouch.SetValue(g_Settings.sz##x, TEXT(#x)); \
                        } \
    // Load reg values, if they don't exist supply defaults
    RegInitDW(DisplayMode, DISPLAYMODE_BOTH);
    RegInitDW(ShowTrailing, 0);
    RegInitDW(ShowData, 1);
    RegInitDW(ShowContact, 0);
    RegInitDW(RawOffsetX, 0);
    RegInitDW(RawOffsetY, (DWORD)-15);      // cast to DWORD to avoid compiler warning
    RegInitDW(GestureOffsetX, 0);
    RegInitDW(GestureOffsetY, (DWORD)-10);
    RegInitDW(ButtonX, BUTTON_HEIGHT);
    RegInitDW(ButtonY, g_dwScreeH - BUTTON_HEIGHT - BUTTON_HEIGHT);
    RegInitDW(WindowX, (g_dwScreeW - DATAWIN_WIDTH) / 2);
    RegInitDW(WindowY, DATAWIN_HEIGHT);
    RegInitDW(BkMode, BKMODE_COLOR);
    RegInitDW(BkColor, RGB(255, 255, 255));
    RegInitDW(DMColor, RGB(0, 0, 255));
    RegInitDW(PANColor, RGB(255, 0, 0));
    RegInitDW(SCROLLColor, RGB(0, 255, 0));
    RegInitDW(SELECTColor, RGB(255, 0, 255));
    RegInitDW(DSELECTColor, RGB(255, 255, 0));
    RegInitDW(HOLDColor, RGB(0, 255, 255));
    RegInitDW(ShowDM, 0);
    RegInitDW(ShowPAN, 1);
    RegInitDW(ShowSCROLL, 1);
    RegInitDW(ShowSELECT, 1);
    RegInitDW(ShowDSELECT, 1);
    RegInitDW(ShowHOLD, 1);
    RegInitDW(ShowStatistics, 0);
    RegInitDW(ShowInitialWarning, 1);
    RegInitDW(ShowIDWarning, 0);
    RegInitDW(ShowContactsWarning, 0);
    RegInitDW(ContactsMin, 12);         // 3 pixels
    RegInitDW(ContactsMax, 80);         // 20 pixels
    RegInitDW(MacroMode, MACROMODE_PLAYBACK);
    RegInitSZ(FileName);
    RegInitSZ(BkImage);

    // macros no longer needed
#undef  RegInitDW
#undef  RegInitSZ

    // Make sure settings are consistent with current filter state
    if (0 == g_fFilterPresent)
    {
        if ((DISPLAYMODE_RAW == g_Settings.dwDisplayMode) || (DISPLAYMODE_BOTH == g_Settings.dwDisplayMode))
        {
            g_Settings.dwDisplayMode = DISPLAYMODE_GESTURE;
        }
        g_Settings.dwShowContact = 0;
    }

    // initialize background bitmap
    if ((BKMODE_IMAGE == g_Settings.dwBkMode) && !UpdateBkBitmap(g_Settings.szBkImage))
    {
        g_Settings.dwBkMode = BKMODE_COLOR;
        LocalFree((HLOCAL)g_Settings.szBkImage);
        g_Settings.szBkImage = StrDup(L"");
    }

}

int
WINAPI
WinMain (
         HINSTANCE  hInstance,
         HINSTANCE  hPrevInstance,
         LPTSTR     lpCmdLine,
         int        nCmdShow
         )
{
    MSG     msg = { NULL, 0, 0,0,0,-1 };
    HRESULT hr;
    HANDLE  hth = NULL;

#ifdef DEBUG_OUTPUT
    NKDbgPrintfW(L"%s started.\r\n", szMainWndTitle);
#endif

    g_hInstance = hInstance;

    s_hEvtAutoData = CreateEvent(NULL, FALSE, FALSE, L"CETouchAutoData");
    if (GetLastError() == ERROR_ALREADY_EXISTS)
    {
        g_fFilterPresent = TRUE;
    }
    else
    {
        CloseHandle(s_hEvtAutoData);
        g_fFilterPresent = FALSE;
    }

    if ( hPrevInstance == 0 )
    {
        LoadSettings();

        if (!g_fFilterPresent && g_Settings.dwShowInitialWarning)
        {
            if (IDNO == MessageBox(NULL, L"CETouchFilter is not loaded.  Run anyway?", L"Initial Warning", MB_YESNO))
            {
                return FALSE;
            }
        }

        if (lpCmdLine && *lpCmdLine)
        {
            if (!g_fFilterPresent)
            {
                MessageBox(NULL, L"CETouchFilter needs to be loaded to use this feature!", L"Filter Not Loaded", MB_OK);
                return FALSE;
            }

            wchar_t *next_token;
            wchar_t *szTok;

            szTok = wcstok_s(lpCmdLine, L" \t", &next_token);
            if ((NULL == szTok) || wcscmp(szTok, L"-t"))
            {
                MessageBox(NULL, L"Unrecognized command line argument!", L"Confuse me you did", MB_OK);
                return FALSE;
            }

            CRegKey hKeyTouch;
            hKeyTouch.Create(HKEY_LOCAL_MACHINE, L"Software\\Microsoft\\CETouchView", NULL, 0, 0, NULL, NULL);

            szTok = wcstok_s(NULL, L" \t", &next_token);
            if (NULL != szTok)
            {
                // szTok should be a file name for playback
                FILE *f = NULL;
                if (0 != _wfopen_s(&f, szTok, L"rt"))
                {
                    MessageBox(NULL, L"Unable to open the file specified.  Nothing to play back.", L"Invalid Filename", MB_OK);
                    return FALSE;
                }
                fclose(f);
                hKeyTouch.SetValue(szTok, L"FileName");
            }
            else
            {
                hKeyTouch.SetValue(L"", L"FileName");
            }

            // set macro mode to playback
            hKeyTouch.SetValue((DWORD)MACROMODE_PLAYBACK, L"MacroMode");

            SetEvent(s_hEvtAutoData);

            return FALSE;
        }

        if ( InitApplication(hInstance) == FALSE )
        {
            return(FALSE);
        }
    }

    if ( InitInstance(hInstance, nCmdShow, &g_hwnd) == FALSE )
    {
        return(FALSE);
    }

    MSGQUEUEOPTIONS opts = {
        sizeof(opts),
        0,
        100,
        sizeof(CETOUCHINPUT[CETOUCHINPUT_MAX_SIMULTANEOUS]),
        TRUE
    };
    s_hMsgQ = CreateMsgQueue(L"TouchInput", &opts);
    CBRA(NULL != s_hMsgQ);

    s_hExit = CreateEvent(NULL, FALSE, FALSE, NULL);
    CBRA(NULL != s_hExit);

    hth = CreateThread(NULL, 0, &PointsThread, NULL, 0, NULL);
    CBR(NULL != hth);

#ifdef DEBUG_OUTPUT
    NKDbgPrintfW(L"%s running.\r\n", szMainWndTitle);
#endif


    InitPerf();

    while (GetMessage(&msg, NULL, 0, 0) != FALSE)
    {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    ExitPerf();

 Error:
    if (hth)
    {
        SetEvent(s_hExit);
        WaitForSingleObject(hth, 30000);
        CloseHandle(hth);
    }

    if (s_hExit) CloseHandle(s_hExit);
    if (s_hMsgQ) CloseHandle(s_hMsgQ);

    for (int i = 0; i < NUM_BRUSHES; i++)
    {
        if (g_rgBrushes[i])
            DeleteObject((HGDIOBJ)g_rgBrushes[i]);
    }

    for (i = 0; i < MAX_ZOOM_LEVEL; i++)
    {
        if (s_hDemoFonts[i])
            DeleteObject((HGDIOBJ)s_hDemoFonts[i]);
    }
    if (s_hBoldFont)
        DeleteObject((HGDIOBJ)s_hBoldFont);

#ifdef DEBUG_OUTPUT
    NKDbgPrintfW(L"%s stopped.\r\n", szMainWndTitle);
#endif

    return(msg.wParam);
}

BOOL InitScreenObjects()
{
    HDC hdc = GetDC(NULL);

    g_rgBrushes[GID_DIRECTMANIPULATION-1] = CreateSolidBrush(g_Settings.dwDMColor);
    g_rgBrushes[GID_PAN-1] = CreateSolidBrush(g_Settings.dwPANColor);
    g_rgBrushes[GID_SCROLL-1] = CreateSolidBrush(g_Settings.dwSCROLLColor);
    g_rgBrushes[GID_SELECT-1] = CreateSolidBrush(g_Settings.dwSELECTColor);
    g_rgBrushes[GID_DOUBLESELECT-1] = CreateSolidBrush(g_Settings.dwDSELECTColor);
    g_rgBrushes[GID_HOLD-1] = CreateSolidBrush(g_Settings.dwHOLDColor);

    g_rgBrushes[BRUSH_BLACK] = CreateSolidBrush(RGB(0,0,0));
    g_rgBrushes[BRUSH_RED] = CreateSolidBrush(RGB(255,0,0));

    HFONT hFont = (HFONT)::GetStockObject(SYSTEM_FONT);
    LOGFONT lf;
    GetObject(hFont, sizeof(LOGFONT), &lf);

    lf.lfQuality = CLEARTYPE_QUALITY;
    lf.lfWeight = FW_BOLD;
    lf.lfHeight = -MulDiv(16, GetDeviceCaps(hdc, LOGPIXELSY), 72);
    s_hBoldFont = ::CreateFontIndirect(&lf);

    int nFontSizeInc = (MAX_FONT_SIZE - MIN_FONT_SIZE) / MAX_ZOOM_LEVEL;
    for (int i = 0; i < MAX_ZOOM_LEVEL; i++)
    {
        lf.lfHeight = -MulDiv(MIN_FONT_SIZE + nFontSizeInc * i,
            GetDeviceCaps(hdc, LOGPIXELSY), 72);
        s_hDemoFonts[i] = ::CreateFontIndirect(&lf);
    }

    if (hdc != NULL)
    {
        ReleaseDC(NULL, hdc);
    }

    return TRUE;
}

void ClearRawList ( void )
{
    ce::gate<ce::critical_section> gate(s_cs);

    while (!RawList.empty())
    {
        LocalFree(RawList.front());
        RawList.pop_front();
    }
}

void DrawRawDots ( void )
{
    BOOL fPostContactsWarning = FALSE;
    BOOL fPostContactsAreaInvalid = FALSE;

    // If nothing to do, simply return
    if (RawList.empty())
    {
        return;
    }

    // If not displaying raw data, empty raw list then return
    if (!((DISPLAYMODE_RAW == g_Settings.dwDisplayMode) || (DISPLAYMODE_BOTH == g_Settings.dwDisplayMode)))
    {
        ClearRawList();
        return;
    }

    ce::gate<ce::critical_section> gate(s_cs);

    HBRUSH hOrgBrush = (HBRUSH)SelectObject(g_hdcPerm, GetStockObject(BLACK_BRUSH));
    s_fIDWarning = FALSE;

    // if we get here, there is something to display and we're supposed to display it
    while (!RawList.empty())
    {
        TouchInputData *td = RawList.front();

        // check for duplicate dwID if enabled and not already found
        if (g_Settings.dwShowIDWarning  && !s_fIDWarning)
        {
            for (size_t i = 1; i < td->cInputs; ++i)
            {
                for (size_t j = 0; j < i; ++j)
                {
                    if (td->rgInputs[i].dwID == td->rgInputs[j].dwID)
                    {
                        s_fIDWarning = TRUE;
                    }
                }
            }
        }

#if 0
        for (size_t i = 0; i < td->cInputs; ++i)
        {
            RETAILMSG(TRUE, (
                L"TP[%d of %d]: (%d,%d) hSource:%08X dwID:%08X dwFlags:%08X dwMask:%08X dwTime:%08X cxContact:%08X cyContact:%08X dwPropertyOffset:%08X cbProperty:%08X\r\n",
                i + 1,
                td->cInputs,
                td->rgInputs[i].x,
                td->rgInputs[i].y,
                td->rgInputs[i].hSource,
                td->rgInputs[i].dwID,
                td->rgInputs[i].dwFlags,
                td->rgInputs[i].dwMask,
                td->rgInputs[i].dwTime,
                td->rgInputs[i].cxContact,
                td->rgInputs[i].cyContact,
                td->rgInputs[i].dwPropertyOffset,
                td->rgInputs[i].cbProperty));
        }
#endif

        for (unsigned int i = 0; i < td->cInputs; ++i)
        {
            CETOUCHINPUT *pInput = &td->rgInputs[i];
            BOOL fContactAreaValid = (0 != (TOUCHINPUTMASKF_CONTACTAREA & pInput->dwMask));

            // touch driver values are in 1/4 pixels, so multiply our offsets by 4; will be divided later
            LONG dx = (LONG)g_Settings.dwRawOffsetX * 4;
            LONG dy = (LONG)g_Settings.dwRawOffsetY * 4;
            LONG r = TOUCHPOINT_RADIUS * 4;
            RECT rcID;

            POINT pt1 = { (pInput->x + dx - r), (pInput->y + dy - r) };
            POINT pt2 = { (pInput->x + dx + r), (pInput->y + dy + r) };


            if (g_Settings.dwShowContact && fContactAreaValid)
            {
                pt1.x -= pInput->cxContact / 2;
                pt2.x += pInput->cxContact / 2;
                pt1.y -= pInput->cyContact / 2;
                pt2.y += pInput->cyContact / 2;
            }

            pt1.x /= 4, pt1.y /= 4;
            pt2.x /= 4, pt2.y /= 4;
            VERIFY(ScreenToClient(g_hwnd, &pt1));
            VERIFY(ScreenToClient(g_hwnd, &pt2));

            if (g_Settings.dwShowContact && fContactAreaValid)
            {
                SetRect(&rcID, pt1.x, pt1.y, pt2.x, pt2.y);
                FillRect(g_hdcPerm, &rcID, (HBRUSH)GetStockObject(BLACK_BRUSH));
            }
            else
            {
                VERIFY(Ellipse(g_hdcPerm, pt1.x, pt1.y, pt2.x, pt2.y));
            }

            if (g_Settings.dwShowContactsWarning &&
                (!fContactAreaValid ||
                (pInput->cxContact < g_Settings.dwContactsMin) ||
                (pInput->cyContact < g_Settings.dwContactsMin) ||
                (pInput->cxContact > g_Settings.dwContactsMax) ||
                (pInput->cyContact > g_Settings.dwContactsMax)))
            {
                fPostContactsWarning = TRUE;
                if (!fContactAreaValid)
                {
                    fPostContactsAreaInvalid = TRUE;
                }
            }

        }

        UpdateDataWindow(NULL, td);
        LocalFree(RawList.front());
        RawList.pop_front();
    }

    SelectObject(g_hdcPerm, (HGDIOBJ)hOrgBrush);
    InvalidateRect(g_hwnd, NULL, FALSE);

    if (fPostContactsWarning)
    {
        PostMessage(g_hwnd, WM_SHOW_CONTACTSWARNING, fPostContactsAreaInvalid, 0);
    }
}

void DrawGestureText ( GESTUREINFO *pgi )
{
    if ((NULL != pgi) && (DISPLAYMODE_TEXT == g_Settings.dwDisplayMode))
    {
        HDC hdc = g_hdcPerm;
        DWORD dwColor = 0;

        HFONT hOrgFont = (HFONT)SelectObject(hdc, (HGDIOBJ)s_hDemoFonts[s_dwZoomLevel - 1]);
        if (s_fUpWarning)
        {
            dwColor = RGB(255, 0, 0);
        }
        else switch (pgi->dwID)
        {
            case GID_BEGIN:
            case GID_END:
                return;
            case GID_DIRECTMANIPULATION:
                dwColor = g_Settings.dwDMColor;
                break;
            case GID_PAN:
                dwColor = g_Settings.dwPANColor;
                break;
            case GID_SCROLL:
                dwColor = g_Settings.dwSCROLLColor;
                break;
            case GID_SELECT:
                dwColor = g_Settings.dwSELECTColor;
                break;
            case GID_DOUBLESELECT:
                dwColor = g_Settings.dwDSELECTColor;
                break;
            case GID_HOLD:
                dwColor = g_Settings.dwHOLDColor;
                break;

        }
        SetTextColor(hdc, dwColor);
        SetBkColor(hdc, g_Settings.dwBkColor);
        DrawText(hdc, L"Sample Text", -1, &s_rcClient, DT_VCENTER|DT_CENTER);
        SelectObject(hdc, (HGDIOBJ)hOrgFont);
    }
    InvalidateRect(g_hwnd, NULL, FALSE);
}

void DrawGestureBar ( GESTUREINFO *pgi )
{
    if (DISPLAYMODE_BAR == g_Settings.dwDisplayMode)
    {
        HDC hdc = g_hdcPerm;
        RECT rc;
        HBRUSH hBrush = NULL;
        DWORD dwRadius = HIDWORD(pgi->ullArguments) / 2;
        if (0 == dwRadius)
        {
            dwRadius = g_dwScreeW / 4;
        }
        dwRadius = min(dwRadius, (DWORD)(g_dwScreeW / 3));

        SetRect(&rc,
            pgi->ptsLocation.x - dwRadius,
            pgi->ptsLocation.y - TOUCHPOINT_RADIUS,
            pgi->ptsLocation.x + dwRadius,
            pgi->ptsLocation.y + TOUCHPOINT_RADIUS);
        rc.left = max(rc.left, s_rcClient.left);
        rc.top = max(rc.top, s_rcClient.top);
        rc.right = min(rc.right, s_rcClient.right);
        rc.bottom = min(rc.bottom, s_rcClient.bottom);
        SetTextColor(hdc, RGB(0,0,0));
        SetBkColor(hdc, g_Settings.dwBkColor);
        DrawEdge(hdc, &rc, EDGE_ETCHED, BF_FLAT | BF_RECT | BF_MONO);
        InflateRect(&rc, -1, -1);
        if (s_fUpWarning)
        {
            hBrush = g_rgBrushes[BRUSH_RED];
        }
        else
        {
            ASSERT(pgi->dwID);
            hBrush = g_rgBrushes[pgi->dwID - 1];
        }
        if (NULL != hBrush)
        {
            FillRect(hdc, &rc, hBrush);
            InvalidateRect(g_hwnd, NULL, FALSE);
        }
    }
}

void DrawGestureDot ( GESTUREINFO *pgi )
{
    BOOL fShowDot = FALSE;

    if (NULL != pgi)
    {
        switch (pgi->dwID)
        {
            case GID_DIRECTMANIPULATION:
                fShowDot = (0 != g_Settings.dwShowDM);
                break;
            case GID_PAN:
                fShowDot = (0 != g_Settings.dwShowPAN);
                break;
            case GID_SCROLL:
                fShowDot = (0 != g_Settings.dwShowSCROLL);
                break;
            case GID_SELECT:
                fShowDot = (0 != g_Settings.dwShowSELECT);
                break;
            case GID_DOUBLESELECT:
                fShowDot = (0 != g_Settings.dwShowDSELECT);
                break;
            case GID_HOLD:
                fShowDot = (0 != g_Settings.dwShowHOLD);
                break;
        }
    }

    if (!fShowDot)
    {
        return;
    }

    POINT pt = { pgi->ptsLocation.x + g_Settings.dwGestureOffsetX, pgi->ptsLocation.y + g_Settings.dwGestureOffsetY};
    DWORD dwID = pgi->dwID;
    HBRUSH hOrgBrush = NULL;

    HDC hDC = g_hdcPerm;
    if (hDC == NULL)
        return;

    if (NULL == g_rgBrushes[dwID-1])
    {
        hOrgBrush = (HBRUSH)SelectObject(hDC, (HGDIOBJ)g_rgBrushes[BRUSH_BLACK]);
    }
    else
    {
        hOrgBrush = (HBRUSH)SelectObject(hDC, (HGDIOBJ)g_rgBrushes[dwID-1]);
    }

    if ((dwID == GID_DIRECTMANIPULATION) && LODWORD(pgi->ullArguments))   // DM multi-touch?
    {
        POINT pt2 = { pt.x + GID_DIRECTMANIPULATION_DELTA_X(pgi->ullArguments), pt.y + GID_DIRECTMANIPULATION_DELTA_Y(pgi->ullArguments) };
        Ellipse(hDC, pt2.x - TOUCHPOINT_RADIUS, pt2.y - TOUCHPOINT_RADIUS, pt2.x + TOUCHPOINT_RADIUS, pt2.y + TOUCHPOINT_RADIUS);
    }
    else if ((dwID == GID_PAN) && ((HIDWORD(pgi->ullArguments)) && !(GF_INERTIA & pgi->dwFlags)))   // !DM multi-touch?
    {
        HBRUSH hBrush = CreateSolidBrush(g_Settings.dwBkColor);
        DWORD dwRadius = GID_PAN_CONTACT_SEPARATION(pgi->ullArguments) >> 1;
        SelectObject(hDC, hBrush);
        Ellipse(hDC, pt.x - dwRadius, pt.y - dwRadius, pt.x + dwRadius, pt.y + dwRadius);
        SelectObject(hDC, (HGDIOBJ)g_rgBrushes[dwID-1]);
        DeleteObject(hBrush);
    }

    Ellipse(hDC, pt.x - TOUCHPOINT_RADIUS, pt.y - TOUCHPOINT_RADIUS, pt.x + TOUCHPOINT_RADIUS, pt.y + TOUCHPOINT_RADIUS);

    SelectObject(hDC, (HGDIOBJ)hOrgBrush);
    InvalidateRect(g_hwnd, NULL, FALSE);
}


BOOL SwapDM (HWND hwnd, BOOL fUseDM)
{
    BOOL fRet = FALSE;

    if (s_fUseDM == fUseDM)
    {
        return TRUE;
    }

    if (fUseDM)
    {
        fRet = DisableGestures(hwnd, TGF_GID_PAN | TGF_GID_SCROLL, TGF_SCOPE_WINDOW);
        if (!fRet)
        {
            OutputDebugString(L"ERROR -- Unable to disable PAN/SCROLL gestures");
        }
        fRet = EnableGestures(hwnd, TGF_GID_DIRECTMANIPULATION, TGF_SCOPE_WINDOW);
        if (!fRet)
        {
            OutputDebugString(L"ERROR -- Unable to enable DM gestures");
        }
        OutputDebugString(L"<<<<< Now using DIRECTMANIPULATION Gestures >>>>>");
    }
    else
    {
        fRet = DisableGestures(hwnd, TGF_GID_DIRECTMANIPULATION, TGF_SCOPE_WINDOW);
        if (!fRet)
        {
            OutputDebugString(L"ERROR -- Unable to disable DM gestures");
        }
        fRet = EnableGestures(hwnd, TGF_GID_PAN | TGF_GID_SCROLL, TGF_SCOPE_WINDOW);
        if (!fRet)
        {
            OutputDebugString(L"ERROR -- Unable to enable PAN/SCROLL gestures");
        }
        OutputDebugString(L">>>>> Now using PAN/SCROLL Gestures <<<<<");
    }

    if (!fRet)
        return FALSE;
    s_fUseDM = fUseDM;

    return TRUE;
}


void ZoomDemo(HWND hWnd, BOOL fZoomIn)
{
    if (fZoomIn && s_dwZoomLevel < MAX_ZOOM_LEVEL)
    {
        s_dwZoomLevel++;
    }
    if (!fZoomIn && s_dwZoomLevel > MIN_ZOOM_LEVEL)
    {
        s_dwZoomLevel--;
    }
}


void PanProc ( GESTUREINFO *pgi )
{
    static DWORD s_dwPrevDistance = 0;
    static BOOL s_bBeginReceived = FALSE;
#if 0
 	NKDbgPrintfW(_T("PanProc: GID=0x%x Flag=0x%x Arg=0x%08x%08x x=%x y=%x\r\n"), 
			pgi->dwID, pgi->dwFlags, HIDWORD(pgi->ullArguments), LODWORD(pgi->ullArguments),
			pgi->ptsLocation.x, pgi->ptsLocation.y);
#endif
    if ((DISPLAYMODE_RAW == g_Settings.dwDisplayMode) ||
        (DISPLAYMODE_GESTURE == g_Settings.dwDisplayMode) ||
        (DISPLAYMODE_BOTH == g_Settings.dwDisplayMode) ||
        (NULL == pgi))
    {
        // nothing to do for these modes
        return;
    }
    else if (!HIDWORD(pgi->ullArguments))
    {
        // single touch, do nothing
        return;
    }
    else if (GF_INERTIA & pgi->dwFlags)    // Check if INERTIA message
    {
        // Ignore these, next message should be a GID_SCROLL
        return;
    }
    else if (GID_DIRECTMANIPULATION == pgi->dwID)
    {
        // DM multi-touch is not handled for these modes
        return;
    }
    else                        // Handle multi-touch gesture
    {
        DWORD dwNewDist = HIDWORD(pgi->ullArguments);
        if (!s_bBeginReceived)
        {
            if (GF_BEGIN & pgi->dwFlags)
            {
                s_bBeginReceived = TRUE;
                s_dwPrevDistance = dwNewDist;
            }
            else
            {
                // error condition - Probably can ignore but break to check msg details
                DebugBreak();
            }
        }
        else
        {
            if (GF_END & pgi->dwFlags)
            {
                s_bBeginReceived = FALSE;
                s_dwPrevDistance = 0;
            }
            else
            {
		   		// multi-touch fingers may have moved, see if we need to zoom
                if (s_dwPrevDistance != dwNewDist)
                {
                    int nDelta = (int)dwNewDist - (int)s_dwPrevDistance;

                    if (g_Settings.dwDisplayMode == DISPLAYMODE_TEXT)
                    {
                        if (nDelta > 15 || nDelta < -15)
                        {
                            ZoomDemo(pgi->hwndTarget, nDelta > 0);
                            s_dwPrevDistance = dwNewDist;
                        }
                    }
                }
            }
        }
    }
}
