//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this sample source code is subject to the terms of the Microsoft
// license agreement under which you licensed this sample source code. If
// you did not accept the terms of the license agreement, you are not
// authorized to use this sample source code. For the terms of the license,
// please see the license agreement between you and Microsoft or, if applicable,
// see the LICENSE.RTF on your install media or the root of your tools installation.
// THE SAMPLE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//

// Contains code to handle the Options Dialog and supporting routines.

#include "CETouchView.h"
#include "regkey.h"

extern REGSETTINGS g_Settings;
extern int g_dwScreeW;
extern int g_dwScreeH;
extern HWND g_hwnd;
extern BOOL g_fFilterPresent;
extern HINSTANCE g_hInstance;
extern BOOL g_fRecording;
extern HWND g_hwndButton;
extern HWND g_hwndData;

int s_iCurrentTab = 0;

const TABINFO rgpTab[] =
{
    { IDS_DISPLAYMODE,      IDD_DISPLAYMODE,    DisplayModeTabDlgProc },
    { IDS_DISPLAYOFFSET,    IDD_DISPLAYOFFSET,  DisplayOffsetTabDlgProc },
    { IDS_USERINTERFACE,    IDD_USERINTERFACE,  UserInterfaceTabDlgProc },
    { IDS_GESTURE,          IDD_GESTURE,        GestureTabDlgProc },
    { IDS_WARNINGS,         IDD_WARNINGS,       WarningsTabDlgProc },
    { IDS_RECPLAY,          IDD_RECPLAY,        RecPlayTabDlgProc },
};

// ValidateNumber: Convert and validate a base 10 number
// Input:  szNum: string to convert
//         cchMax: max length of szNum
//         iMin: minimum valid value
//         iMax: maximum valid value
//         *pdwNum: the converted number, can be NULL if only valildation is required
//
// returns TRUE if valid, FALSE if invalid.
BOOL ValidateNumber ( LPCTSTR szNum, size_t cchMax, int iMin, int iMax, DWORD *pdwNum )
{
    size_t iLen = 0;
    int iNum = 0;
    BOOL fNeg = FALSE;

    if (FAILED(StringCchLength(szNum, cchMax, &iLen)))
    {
        return FALSE;
    }

    for (size_t i = 0; i < iLen; ++i)
    {
        WCHAR c = szNum[i];

        // converts 0..9
        if ((L'0' <= c) && (L'9' >= c))
        {
            iNum = (iNum * 10) + (c - L'0');
        }
        // check for minus sign only in first position
        else if ((0 == i) && (L'-' == c))
        {
            fNeg = TRUE;
        }
        // ignore plus sign only in first position
        else if ((0 == i) && (L'+' == c))
        {
            // allow a + sign but ignore it
            ;
        }
        // anything else invalidates the number
        else
        {
            return FALSE;
        }

    }

    // negate if we found the minus sign
    if (fNeg)
    {
        iNum = -iNum;
    }

    // make sure in proper range
    if ((iNum < iMin) || (iNum > iMax))
    {
        return FALSE;
    }

    // return value if requested
    if (NULL != pdwNum)
    {
        *pdwNum = (DWORD)iNum;
    }
    return TRUE;
}

// ValidateColor: Convert and validate a hex string to a color value
// Input:  szColor: string to convert
//         cchMax: max length of szColor
//         *pdwColor: the converted color, can be NULL if only valildation is required
//
// returns TRUE if valid, FALSE if invalid.
BOOL ValidateColor ( LPCTSTR szColor, size_t cchMax, DWORD *pdwColor )
{
    size_t iLen = 0;
    DWORD dwColor = 0;

    if (FAILED(StringCchLength(szColor, cchMax, &iLen)))
    {
        return FALSE;
    }

    for (size_t i = 0; i < iLen; ++i)
    {
        WCHAR c = towupper(szColor[i]);
        // converts 0..9
        if ((L'0' <= c) && (L'9' >= c))
        {
            dwColor = (dwColor << 4) + (c - L'0');
        }
        // converts A..F
        else if ((L'A' <= c) && (L'F' >= c))
        {
            dwColor = (dwColor << 4) + (c - L'A') + 10;
        }
        // anything else invalidates the number
        else
        {
            return FALSE;
        }

        // Since we're checking at each nibble, this won't overflow
        if (0xFFFFFF < dwColor)
        {
            return FALSE;
        }
    }

    // return value if requested
    if (NULL != pdwColor)
    {
        *pdwColor = dwColor;
    }
    return TRUE;
}

// Load a string resource, allocate a buffer exactly large enough for it and the terminating null.
LPCTSTR LoadStringZ ( UINT id )
{
    WORD nLen;
    LPTSTR szCopy;
    LPCTSTR szRes = (LPCTSTR)LoadString(g_hInstance, id, NULL, 0);

    nLen = *(WORD *)(szRes - 1);
    szCopy = (LPTSTR)LocalAlloc(LPTR, (nLen + 1) * sizeof(WCHAR));
    StringCchCopyN(szCopy, nLen + 1, szRes, nLen);

    return szCopy;
}

// Main processing of options dialog.  Sets up property sheets, and updates registry and program
// settings as necessary
void OptionsDialog ( HWND hwndParent )
{
    PROPSHEETPAGE   *rgPropPages = NULL;
    PROPSHEETHEADER PropHdr = {0};
    REGSETTINGS rsCopy = {0};

    // Set up property pages
#pragma prefast(suppress:6260, "use of multiplied sizeof's is correct here")
    rgPropPages = (PROPSHEETPAGE *) LocalAlloc(LPTR, _countof(rgpTab) * sizeof(PROPSHEETPAGE));
    if (NULL == rgPropPages)
    {
        // Error!  Out of memory...invoke debugger for the moment...
        DebugBreak();
        return;
    }

    // Copy registry settings
    rsCopy = g_Settings;
    rsCopy.szFileName = StrDup(g_Settings.szFileName);
    rsCopy.szBkImage = StrDup(g_Settings.szBkImage);

    for (int iPage = 0; iPage < _countof(rgpTab); ++iPage)
    {
        // Don't create last tab if filter is not installed
        if (!g_fFilterPresent && (iPage == (_countof(rgpTab) - 1)))
        {
            continue;
        }
        rgPropPages[iPage].dwSize = sizeof(PROPSHEETPAGE);
        rgPropPages[iPage].dwFlags = PSP_USETITLE;
        rgPropPages[iPage].hInstance = g_hInstance;
        rgPropPages[iPage].pszTemplate = MAKEINTRESOURCE(rgpTab[iPage].iddDlg);
        rgPropPages[iPage].pszIcon = NULL;
        rgPropPages[iPage].pfnDlgProc = rgpTab[iPage].pfnDlg;
        rgPropPages[iPage].pszTitle = LoadStringZ(rgpTab[iPage].idsTitle);
        rgPropPages[iPage].pfnCallback = NULL;
        rgPropPages[iPage].lParam = (LPARAM)&rsCopy;
    }

    // Set up property sheet header
    PropHdr.dwSize = sizeof(PROPSHEETHEADER);
    PropHdr.dwFlags = PSH_PROPSHEETPAGE;
    PropHdr.hwndParent = hwndParent;
    PropHdr.hInstance = g_hInstance;
    PropHdr.pszCaption = LoadStringZ(IDS_OPTIONS);
    PropHdr.nPages = _countof(rgpTab) - (g_fFilterPresent ? 0 : 1);
    PropHdr.ppsp = rgPropPages;
    PropHdr.nStartPage = s_iCurrentTab;

    // Invoke property sheet
    if (IDOK == PropertySheet(&PropHdr))
    {
        CRegKey hKeyTouch;

        // Create reg key if it doesn't exist, or open it if it does:
        hKeyTouch.Create(HKEY_LOCAL_MACHINE, L"Software\\Microsoft\\CETouchView", NULL, 0, 0, NULL, NULL);
    // Use these nifty macros to make saving all the reg values more concise
#define RegSaveDW(x)    if (g_Settings.dw##x != rsCopy.dw##x) \
                        { \
                            hKeyTouch.SetValue(g_Settings.dw##x = rsCopy.dw##x, TEXT(#x)); \
                        }
#define RegSaveSZ(x)    if (0 != wcscmp(g_Settings.sz##x,rsCopy.sz##x)) \
                        { \
                            if (NULL != g_Settings.sz##x) \
                            { \
                                LocalFree((HLOCAL)g_Settings.sz##x); \
                            } \
                            if (NULL != rsCopy.sz##x) \
                            { \
                                hKeyTouch.SetValue(g_Settings.sz##x = StrDup(rsCopy.sz##x), TEXT(#x)); \
                            } \
                            else \
                            { \
                                g_Settings.sz##x = NULL; \
                                hKeyTouch.SetValue(L"", TEXT(#x)); \
                            } \
                        }
        // Save reg values if they were changed
        RegSaveDW(DisplayMode);
        RegSaveDW(ShowTrailing);
        RegSaveDW(ShowData);
        RegSaveDW(ShowContact);
        RegSaveDW(RawOffsetX);
        RegSaveDW(RawOffsetY);
        RegSaveDW(GestureOffsetX);
        RegSaveDW(GestureOffsetY);
        RegSaveDW(ButtonX);
        RegSaveDW(ButtonY);
        RegSaveDW(WindowX);
        RegSaveDW(WindowY);
        RegSaveDW(BkMode);
        RegSaveDW(BkColor);
        RegSaveDW(DMColor);
        RegSaveDW(PANColor);
        RegSaveDW(SCROLLColor);
        RegSaveDW(SELECTColor);
        RegSaveDW(DSELECTColor);
        RegSaveDW(HOLDColor);
        RegSaveDW(ShowDM);
        RegSaveDW(ShowPAN);
        RegSaveDW(ShowSCROLL);
        RegSaveDW(ShowSELECT);
        RegSaveDW(ShowDSELECT);
        RegSaveDW(ShowHOLD);
        RegSaveDW(ShowStatistics);
        RegSaveDW(ShowInitialWarning);
        RegSaveDW(ShowIDWarning);
        RegSaveDW(ShowContactsWarning);
        RegSaveDW(ContactsMin);
        RegSaveDW(ContactsMax);
        RegSaveDW(MacroMode);
        RegSaveSZ(FileName);
        RegSaveSZ(BkImage);

#undef RegSaveDW
#undef RegSaveSZ
    }

    // Move Screen Elements to new locations
    MoveWindow(g_hwndButton, g_Settings.dwButtonX, g_Settings.dwButtonY, BUTTON_WIDTH, BUTTON_HEIGHT, FALSE);
    MoveWindow(g_hwndData, g_Settings.dwWindowX, g_Settings.dwWindowY, DATAWIN_WIDTH, DATAWIN_HEIGHT, FALSE);

    // Show or hide data window
    ShowWindow(g_hwndData, 0 == g_Settings.dwShowData ? SW_HIDE : SW_SHOW);

    // reload background bitmap
    if ((BKMODE_IMAGE == g_Settings.dwBkMode) && !UpdateBkBitmap(g_Settings.szBkImage))
    {
        g_Settings.dwBkMode = BKMODE_COLOR;
        LocalFree((HLOCAL)g_Settings.szBkImage);
        g_Settings.szBkImage = StrDup(L"");
    }

    // reload gesture brushes with new colors
    DeleteObject(g_rgBrushes[GID_DIRECTMANIPULATION-1]);
    g_rgBrushes[GID_DIRECTMANIPULATION-1] = CreateSolidBrush(g_Settings.dwDMColor);
    DeleteObject(g_rgBrushes[GID_PAN-1]);
    g_rgBrushes[GID_PAN-1] = CreateSolidBrush(g_Settings.dwPANColor);
    DeleteObject(g_rgBrushes[GID_SCROLL-1]);
    g_rgBrushes[GID_SCROLL-1] = CreateSolidBrush(g_Settings.dwSCROLLColor);
    DeleteObject(g_rgBrushes[GID_SELECT-1]);
    g_rgBrushes[GID_SELECT-1] = CreateSolidBrush(g_Settings.dwSELECTColor);
    DeleteObject(g_rgBrushes[GID_DOUBLESELECT-1]);
    g_rgBrushes[GID_DOUBLESELECT-1] = CreateSolidBrush(g_Settings.dwDSELECTColor);
    DeleteObject(g_rgBrushes[GID_HOLD-1]);
    g_rgBrushes[GID_HOLD-1] = CreateSolidBrush(g_Settings.dwHOLDColor);

    // Refresh Main Window
    ClearScreen();

    // Clear out any accumulated raw data points while options were displayed
    ClearRawList();

    // Free allocated mem...
    if (NULL != rgPropPages)
    {
        for (int iPage = 0; iPage < _countof(rgpTab); ++iPage)
        {
            LocalFree((HLOCAL)rgPropPages[iPage].pszTitle);
        }
        LocalFree(rgPropPages);
    }
    if (NULL != PropHdr.pszCaption)
    {
        LocalFree((HLOCAL)PropHdr.pszCaption);
    }
    LocalFree((HLOCAL)rsCopy.szFileName);
    LocalFree((HLOCAL)rsCopy.szBkImage);
}

// DlgProc for Display Mode tab
BOOL DisplayModeTabDlgProc ( HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    PROPSHEETPAGE *pSheet = NULL;
    static REGSETTINGS *pSettings = NULL;

    switch (uMsg)
    {
        case WM_INITDIALOG:
            ASSERT(NULL != lParam);
            pSheet = (PROPSHEETPAGE *)lParam;
            ASSERT(NULL != pSheet->lParam);
            pSettings = (REGSETTINGS *)pSheet->lParam;

            // Disable appropriate controls if filter is not installed
            if (0 == g_fFilterPresent)
            {
                // These should have been disallowed in LoadSettings()
                ASSERT(DISPLAYMODE_RAW != pSettings->dwDisplayMode);
                ASSERT(DISPLAYMODE_BOTH != pSettings->dwDisplayMode);
                ASSERT(0 == pSettings->dwShowContact);

                EnableWindow(GetDlgItem(hDlg, IDC_DISPLAYMODE_RAW), FALSE);
                EnableWindow(GetDlgItem(hDlg, IDC_DISPLAYMODE_BOTH), FALSE);
                EnableWindow(GetDlgItem(hDlg, IDC_DISPLAYMODE_CONTACT), FALSE);
            }

            // Disable appropriate controls depending on current display mode
            if ((DISPLAYMODE_TEXT == pSettings->dwDisplayMode) ||
                (DISPLAYMODE_BAR == pSettings->dwDisplayMode) ||
                (DISPLAYMODE_GESTURE == pSettings->dwDisplayMode))
            {
                EnableWindow(GetDlgItem(hDlg, IDC_DISPLAYMODE_CONTACT), FALSE);
            }

            SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_RAW, BM_SETCHECK, DISPLAYMODE_RAW == pSettings->dwDisplayMode, 0);
            SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_GESTURE, BM_SETCHECK, DISPLAYMODE_GESTURE == pSettings->dwDisplayMode, 0);
            SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_BOTH, BM_SETCHECK, DISPLAYMODE_BOTH == pSettings->dwDisplayMode, 0);
            SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_TEXT, BM_SETCHECK, DISPLAYMODE_TEXT == pSettings->dwDisplayMode, 0);
            SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_BAR, BM_SETCHECK, DISPLAYMODE_BAR == pSettings->dwDisplayMode, 0);

            SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_TRAILING, BM_SETCHECK, pSettings->dwShowTrailing, 0);
            SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_CONTACT, BM_SETCHECK, pSettings->dwShowContact, 0);
            SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_DATA, BM_SETCHECK, pSettings->dwShowData, 0);
            SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_STATISTICS, BM_SETCHECK, pSettings->dwShowStatistics, 0);
            break;
        case WM_DESTROY:
            break;
        case WM_COMMAND:
            if (IDB_EXIT_APP == wParam)
            {
                if (IDYES == MessageBox(g_hwnd, L"Are you sure you want to exit CETouchView?", L"Exit Application", MB_YESNO))
                {
                    DestroyWindow(g_hwnd);
                }
            }
            else if (IDB_OK == wParam)
            {
                // Tell the property sheet to save changes
                PostMessage(GetParent(hDlg), WM_COMMAND, IDOK, NULL);
            }
            else if (IDB_CANCEL == wParam)
            {
                // Tell the property sheet to cancel changes
                PostMessage(GetParent(hDlg), WM_COMMAND, IDCANCEL, NULL);
            }
            else if (IDB_NEXT_TAB == wParam)
            {
                SendMessage(GetParent(hDlg), PSM_SETCURSEL, (s_iCurrentTab + 1) % (g_fFilterPresent ? _countof(rgpTab) : _countof(rgpTab) - 1), NULL);
            }
            else if (IDB_PREV_TAB == wParam)
            {
                SendMessage(GetParent(hDlg),
                    PSM_SETCURSEL,
                    (s_iCurrentTab ? s_iCurrentTab - 1 : (g_fFilterPresent ? _countof(rgpTab) - 1 : _countof(rgpTab) - 2)),
                    NULL);
            }
            else if (BN_CLICKED == HIWORD(wParam))
            {
                switch (LOWORD(wParam))
                {
                    case IDC_DISPLAYMODE_DATA:
                        if (BST_UNCHECKED == SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_DATA, BM_GETCHECK, 0, 0))
                        {
                            MessageBox(hDlg, L"By disabling the Data Window, you also will not see warning messages or statistics data, if enabled.", L"Disabling Data Window", MB_OK);
                        }
                        break;
                    case IDC_DISPLAYMODE_GESTURE:
                    case IDC_DISPLAYMODE_RAW:
                    case IDC_DISPLAYMODE_BOTH:
                        // Contact is enabled/disabled if filter is installed
                        if (g_fFilterPresent)
                        {
                            EnableWindow(GetDlgItem(hDlg, IDC_DISPLAYMODE_CONTACT), (IDC_DISPLAYMODE_GESTURE != LOWORD(wParam)));
                        }
                        break;
                    case IDC_DISPLAYMODE_TEXT:
                    case IDC_DISPLAYMODE_BAR:
                        // If text or bar, contact is disabled
                        EnableWindow(GetDlgItem(hDlg, IDC_DISPLAYMODE_CONTACT), FALSE);
                        break;
                }
            }
            break;
        case WM_NOTIFY:
            switch (((LPNMHDR)lParam)->code)
            {
                case PSN_SETACTIVE:
                    s_iCurrentTab = 0;
                    break;
                case PSN_APPLY:
                    if (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_RAW, BM_GETCHECK, 0, 0))
                    {
                        pSettings->dwDisplayMode = DISPLAYMODE_RAW;
                    }
                    else if (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_GESTURE, BM_GETCHECK, 0, 0))
                    {
                        pSettings->dwDisplayMode = DISPLAYMODE_GESTURE;
                    }
                    else if (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_BOTH, BM_GETCHECK, 0, 0))
                    {
                        pSettings->dwDisplayMode = DISPLAYMODE_BOTH;
                    }
                    else if (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_TEXT, BM_GETCHECK, 0, 0))
                    {
                        pSettings->dwDisplayMode = DISPLAYMODE_TEXT;
                    }
                    else if (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_BAR, BM_GETCHECK, 0, 0))
                    {
                        pSettings->dwDisplayMode = DISPLAYMODE_BAR;
                    }
                    else
                    {
                        // error, this should never happen
                        ASSERT(FALSE);
                    }

                    pSettings->dwShowTrailing = (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_TRAILING, BM_GETCHECK, 0, 0));
                    pSettings->dwShowContact = (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_CONTACT, BM_GETCHECK, 0, 0));
                    pSettings->dwShowData = (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_DATA, BM_GETCHECK, 0, 0));
                    pSettings->dwShowStatistics = (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_DISPLAYMODE_STATISTICS, BM_GETCHECK, 0, 0));

                    // if display mode is text or bar, contact should be clear
                    if ((DISPLAYMODE_TEXT == pSettings->dwDisplayMode) || (DISPLAYMODE_BAR == pSettings->dwDisplayMode))
                    {
                        pSettings->dwShowContact = 0;
                    }
                    break;
            }
            break;
    }

    return FALSE;
}

// DlgProc for Display Offset tab
BOOL DisplayOffsetTabDlgProc ( HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    WCHAR szTmp[64];
    PROPSHEETPAGE *pSheet = NULL;
    static REGSETTINGS *pSettings = NULL;

    switch (uMsg)
    {
        case WM_INITDIALOG:
            ASSERT(NULL != lParam);
            pSheet = (PROPSHEETPAGE *)lParam;
            ASSERT(NULL != pSheet->lParam);
            pSettings = (REGSETTINGS *)pSheet->lParam;

            StringCchPrintf(szTmp, _countof(szTmp), L"%d", pSettings->dwRawOffsetX);
            SetDlgItemText(hDlg, IDC_DISPLAYOFFSET_RAWX, szTmp);
            StringCchPrintf(szTmp, _countof(szTmp), L"%d", pSettings->dwRawOffsetY);
            SetDlgItemText(hDlg, IDC_DISPLAYOFFSET_RAWY, szTmp);
            StringCchPrintf(szTmp, _countof(szTmp), L"%d", pSettings->dwGestureOffsetX);
            SetDlgItemText(hDlg, IDC_DISPLAYOFFSET_GESTUREX, szTmp);
            StringCchPrintf(szTmp, _countof(szTmp), L"%d", pSettings->dwGestureOffsetY);
            SetDlgItemText(hDlg, IDC_DISPLAYOFFSET_GESTUREY, szTmp);

            if (0 == g_fFilterPresent)
            {
                EnableWindow(GetDlgItem(hDlg, IDC_DISPLAYOFFSET_RAWX), FALSE);
                EnableWindow(GetDlgItem(hDlg, IDC_DISPLAYOFFSET_RAWY), FALSE);
                EnableWindow(GetDlgItem(hDlg, IDC_DISPLAYOFFSET_RAWTEXT), FALSE);
                EnableWindow(GetDlgItem(hDlg, IDC_DISPLAYOFFSET_XTEXT), FALSE);
                EnableWindow(GetDlgItem(hDlg, IDC_DISPLAYOFFSET_YTEXT), FALSE);
                SetFocus(GetDlgItem(hDlg, IDC_DISPLAYOFFSET_GESTUREX));
            }

            break;
        case WM_DESTROY:
            break;
        case WM_COMMAND:
            if (IDB_EXIT_APP == wParam)
            {
                if (IDYES == MessageBox(g_hwnd, L"Are you sure you want to exit CETouchView?", L"Exit Application", MB_YESNO))
                {
                    DestroyWindow(g_hwnd);
                }
            }
            else if (IDB_OK == wParam)
            {
                // Tell the property sheet to save changes
                PostMessage(GetParent(hDlg), WM_COMMAND, IDOK, NULL);
            }
            else if (IDB_CANCEL == wParam)
            {
                // Tell the property sheet to cancel changes
                PostMessage(GetParent(hDlg), WM_COMMAND, IDCANCEL, NULL);
            }
            else if (IDB_NEXT_TAB == wParam)
            {
                SendMessage(GetParent(hDlg), PSM_SETCURSEL, (s_iCurrentTab + 1) % (g_fFilterPresent ? _countof(rgpTab) : _countof(rgpTab) - 1), NULL);
            }
            else if (IDB_PREV_TAB == wParam)
            {
                SendMessage(GetParent(hDlg),
                    PSM_SETCURSEL,
                    (s_iCurrentTab ? s_iCurrentTab - 1 : (g_fFilterPresent ? _countof(rgpTab) - 1 : _countof(rgpTab) - 2)),
                    NULL);
            }
            break;
        case WM_NOTIFY:
            switch (((LPNMHDR)lParam)->code)
            {
                case PSN_SETACTIVE:
                    s_iCurrentTab = 1;
                    break;
                case PSN_APPLY:
                    GetDlgItemText(hDlg, IDC_DISPLAYOFFSET_RAWX, szTmp, _countof(szTmp));
                    if (!ValidateNumber(szTmp, _countof(szTmp), -50, 50, &pSettings->dwRawOffsetX))
                    {
                        MessageBox(hDlg, L"Raw X is invalid.  Please specify a valid number (-50..50).", L"Invalid offset", MB_OK);
                        SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID);
                        return TRUE;
                    }
                    GetDlgItemText(hDlg, IDC_DISPLAYOFFSET_RAWY, szTmp, _countof(szTmp));
                    if (!ValidateNumber(szTmp, _countof(szTmp), -50, 50, &pSettings->dwRawOffsetY))
                    {
                        MessageBox(hDlg, L"Raw Y is invalid.  Please specify a valid number (-50..50).", L"Invalid offset", MB_OK);
                        SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID);
                        return TRUE;
                    }
                    GetDlgItemText(hDlg, IDC_DISPLAYOFFSET_GESTUREX, szTmp, _countof(szTmp));
                    if (!ValidateNumber(szTmp, _countof(szTmp), -50, 50, &pSettings->dwGestureOffsetX))
                    {
                        MessageBox(hDlg, L"Gesture X is invalid.  Please specify a valid number (-50..50).", L"Invalid offset", MB_OK);
                        SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID);
                        return TRUE;
                    }
                    GetDlgItemText(hDlg, IDC_DISPLAYOFFSET_GESTUREY, szTmp, _countof(szTmp));
                    if (!ValidateNumber(szTmp, _countof(szTmp), -50, 50, &pSettings->dwGestureOffsetY))
                    {
                        MessageBox(hDlg, L"Gesture Y is invalid.  Please specify a valid number (-50..50).", L"Invalid offset", MB_OK);
                        SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID);
                        return TRUE;
                    }
                    break;
            }
            break;
    }

    return FALSE;
}

// DlgProc for User Interface tab
BOOL UserInterfaceTabDlgProc ( HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    WCHAR szTmp[64];
    PROPSHEETPAGE *pSheet = NULL;
    static REGSETTINGS *pSettings = NULL;

    switch (uMsg)
    {
        case WM_INITDIALOG:
            ASSERT(NULL != lParam);
            pSheet = (PROPSHEETPAGE *)lParam;
            ASSERT(NULL != pSheet->lParam);
            pSettings = (REGSETTINGS *)pSheet->lParam;

            StringCchPrintf(szTmp, _countof(szTmp), L"%d", pSettings->dwButtonX);
            SetDlgItemText(hDlg, IDC_USERINTERFACE_BUTTONX, szTmp);
            StringCchPrintf(szTmp, _countof(szTmp), L"%d", pSettings->dwButtonY);
            SetDlgItemText(hDlg, IDC_USERINTERFACE_BUTTONY, szTmp);
            StringCchPrintf(szTmp, _countof(szTmp), L"%d", pSettings->dwWindowX);
            SetDlgItemText(hDlg, IDC_USERINTERFACE_WINDOWX, szTmp);
            StringCchPrintf(szTmp, _countof(szTmp), L"%d", pSettings->dwWindowY);
            SetDlgItemText(hDlg, IDC_USERINTERFACE_WINDOWY, szTmp);

            SendDlgItemMessage(hDlg, IDC_USERINTERFACE_IMAGE, BM_SETCHECK, 1 == pSettings->dwBkMode, 0);
            SendDlgItemMessage(hDlg, IDC_USERINTERFACE_COLOR, BM_SETCHECK, 0 == pSettings->dwBkMode, 0);
            SendDlgItemMessage(hDlg, IDC_USERINTERFACE_NONE, BM_SETCHECK, FALSE, 0);

            SetDlgItemText(hDlg, IDC_USERINTERFACE_IMAGENAME, pSettings->szBkImage);
            StringCchPrintf(szTmp, _countof(szTmp), L"%06X", pSettings->dwBkColor);
            SetDlgItemText(hDlg, IDC_USERINTERFACE_COLORVAL, szTmp);

            break;
        case WM_DESTROY:
            break;
        case WM_COMMAND:
            if (IDB_EXIT_APP == wParam)
            {
                if (IDYES == MessageBox(g_hwnd, L"Are you sure you want to exit CETouchView?", L"Exit Application", MB_YESNO))
                {
                    DestroyWindow(g_hwnd);
                }
            }
            else if (IDB_OK == wParam)
            {
                // Tell the property sheet to save changes
                PostMessage(GetParent(hDlg), WM_COMMAND, IDOK, NULL);
            }
            else if (IDB_CANCEL == wParam)
            {
                // Tell the property sheet to cancel changes
                PostMessage(GetParent(hDlg), WM_COMMAND, IDCANCEL, NULL);
            }
            else if (IDB_NEXT_TAB == wParam)
            {
                SendMessage(GetParent(hDlg), PSM_SETCURSEL, (s_iCurrentTab + 1) % (g_fFilterPresent ? _countof(rgpTab) : _countof(rgpTab) - 1), NULL);
            }
            else if (IDB_PREV_TAB == wParam)
            {
                SendMessage(GetParent(hDlg),
                    PSM_SETCURSEL,
                    (s_iCurrentTab ? s_iCurrentTab - 1 : (g_fFilterPresent ? _countof(rgpTab) - 1 : _countof(rgpTab) - 2)),
                    NULL);
            }
            break;
        case WM_NOTIFY:
            switch (((LPNMHDR)lParam)->code)
            {
                case PSN_SETACTIVE:
                    s_iCurrentTab = 2;
                    break;
                case PSN_APPLY:
                    GetDlgItemText(hDlg, IDC_USERINTERFACE_BUTTONX, szTmp, _countof(szTmp));
                    if (!ValidateNumber(szTmp, _countof(szTmp), 0, g_dwScreeW - BUTTON_WIDTH, &pSettings->dwButtonX))
                    {
                        StringCchPrintf(szTmp, _countof(szTmp), L"Button X is invalid.  Please specify a valid number (0..%d).", g_dwScreeW - BUTTON_WIDTH);
                        MessageBox(hDlg, szTmp, L"Invalid Button Location", MB_OK);
                        SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID);
                        return TRUE;
                    }
                    GetDlgItemText(hDlg, IDC_USERINTERFACE_BUTTONY, szTmp, _countof(szTmp));
                    if (!ValidateNumber(szTmp, _countof(szTmp), 0, g_dwScreeH - BUTTON_HEIGHT, &pSettings->dwButtonY))
                    {
                        StringCchPrintf(szTmp, _countof(szTmp), L"Button Y is invalid.  Please specify a valid number (0..%d).", g_dwScreeH - BUTTON_HEIGHT);
                        MessageBox(hDlg, szTmp, L"Invalid Button Location", MB_OK);
                        SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID);
                        return TRUE;
                    }
                    GetDlgItemText(hDlg, IDC_USERINTERFACE_WINDOWX, szTmp, _countof(szTmp));
                    if (!ValidateNumber(szTmp, _countof(szTmp), 0, g_dwScreeW - DATAWIN_WIDTH, &pSettings->dwWindowX))
                    {
                        StringCchPrintf(szTmp, _countof(szTmp), L"Window X is invalid.  Please specify a valid number (0..%d).", g_dwScreeW - DATAWIN_WIDTH);
                        MessageBox(hDlg, szTmp, L"Invalid Window Location", MB_OK);
                        SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID);
                        return TRUE;
                    }
                    GetDlgItemText(hDlg, IDC_USERINTERFACE_WINDOWY, szTmp, _countof(szTmp));
                    if (!ValidateNumber(szTmp, _countof(szTmp), 0, g_dwScreeH - DATAWIN_HEIGHT, &pSettings->dwWindowY))
                    {
                        StringCchPrintf(szTmp, _countof(szTmp), L"Button Y is invalid.  Please specify a valid number (0..%d).", g_dwScreeH - DATAWIN_HEIGHT);
                        MessageBox(hDlg, szTmp, L"Invalid Window Location", MB_OK);
                        SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID);
                        return TRUE;
                    }

                    GetDlgItemText(hDlg, IDC_USERINTERFACE_IMAGENAME, szTmp, _countof(szTmp));
                    if (wcscmp(szTmp, pSettings->szBkImage))
                    {
                        LocalFree((HLOCAL)pSettings->szBkImage);
                        pSettings->szBkImage = StrDup(szTmp);
                    }
                    GetDlgItemText(hDlg, IDC_USERINTERFACE_COLORVAL, szTmp, _countof(szTmp));
                    if (!ValidateColor(szTmp, _countof(szTmp), &pSettings->dwBkColor))
                    {
                        MessageBox(hDlg, L"Background Color is invalid.  Please specify a valid color (BBGGRR).", L"Invalid Color", MB_OK);
                        SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID);
                        return TRUE;
                    }

                    if (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_USERINTERFACE_IMAGE, BM_GETCHECK, 0, 0))
                    {
                        pSettings->dwBkMode = BKMODE_IMAGE;
                    }
                    else if (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_USERINTERFACE_COLOR, BM_GETCHECK, 0, 0))
                    {
                        pSettings->dwBkMode = BKMODE_COLOR;
                    }
                    else if (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_USERINTERFACE_NONE, BM_GETCHECK, 0, 0))
                    {
                        pSettings->dwBkMode = BKMODE_COLOR;
                        pSettings->dwBkColor = RGB(255, 255, 255);
                    }
                    else
                    {
                        // error, this should never happen
                        ASSERT(FALSE);
                    }

                    // Now once everything is updated, insure that if image is selected, name is valid
                    if ((BKMODE_IMAGE == pSettings->dwBkMode) && !UpdateBkBitmap(pSettings->szBkImage))
                    {
                        MessageBox(hDlg, L"Image is selected, but file specified is not a valid image file.  Specify a valid image or choose Color background.", L"Invalid File Name", MB_OK);
                        SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID);
                        return TRUE;
                    }

                    break;
            }
            break;
    }

    return FALSE;
}

// DlgProc for Gesture tab
BOOL GestureTabDlgProc ( HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    WCHAR szTmp[64];
    PROPSHEETPAGE *pSheet = NULL;
    static REGSETTINGS *pSettings = NULL;

    switch (uMsg)
    {
        case WM_INITDIALOG:
            ASSERT(NULL != lParam);
            pSheet = (PROPSHEETPAGE *)lParam;
            ASSERT(NULL != pSheet->lParam);
            pSettings = (REGSETTINGS *)pSheet->lParam;

            SendDlgItemMessage(hDlg, IDC_GESTURE_DM, BM_SETCHECK, pSettings->dwShowDM, 0);
            SendDlgItemMessage(hDlg, IDC_GESTURE_PAN, BM_SETCHECK, pSettings->dwShowPAN, 0);
            SendDlgItemMessage(hDlg, IDC_GESTURE_SCROLL, BM_SETCHECK, pSettings->dwShowSCROLL, 0);
            SendDlgItemMessage(hDlg, IDC_GESTURE_SELECT, BM_SETCHECK, pSettings->dwShowSELECT, 0);
            SendDlgItemMessage(hDlg, IDC_GESTURE_DSELECT, BM_SETCHECK, pSettings->dwShowDSELECT, 0);
            SendDlgItemMessage(hDlg, IDC_GESTURE_HOLD, BM_SETCHECK, pSettings->dwShowHOLD, 0);

            StringCchPrintf(szTmp, _countof(szTmp), L"%06X", pSettings->dwDMColor);
            SetDlgItemText(hDlg, IDC_GESTURE_DM_COLOR, szTmp);
            StringCchPrintf(szTmp, _countof(szTmp), L"%06X", pSettings->dwPANColor);
            SetDlgItemText(hDlg, IDC_GESTURE_PAN_COLOR, szTmp);
            StringCchPrintf(szTmp, _countof(szTmp), L"%06X", pSettings->dwSCROLLColor);
            SetDlgItemText(hDlg, IDC_GESTURE_SCROLL_COLOR, szTmp);
            StringCchPrintf(szTmp, _countof(szTmp), L"%06X", pSettings->dwSELECTColor);
            SetDlgItemText(hDlg, IDC_GESTURE_SELECT_COLOR, szTmp);
            StringCchPrintf(szTmp, _countof(szTmp), L"%06X", pSettings->dwDSELECTColor);
            SetDlgItemText(hDlg, IDC_GESTURE_DSELECT_COLOR, szTmp);
            StringCchPrintf(szTmp, _countof(szTmp), L"%06X", pSettings->dwHOLDColor);
            SetDlgItemText(hDlg, IDC_GESTURE_HOLD_COLOR, szTmp);

            break;
        case WM_DESTROY:
            break;
        case WM_COMMAND:
            if (IDB_EXIT_APP == wParam)
            {
                if (IDYES == MessageBox(g_hwnd, L"Are you sure you want to exit CETouchView?", L"Exit Application", MB_YESNO))
                {
                    DestroyWindow(g_hwnd);
                }
            }
            else if (IDB_OK == wParam)
            {
                // Tell the property sheet to save changes
                PostMessage(GetParent(hDlg), WM_COMMAND, IDOK, NULL);
            }
            else if (IDB_CANCEL == wParam)
            {
                // Tell the property sheet to cancel changes
                PostMessage(GetParent(hDlg), WM_COMMAND, IDCANCEL, NULL);
            }
            else if (IDB_NEXT_TAB == wParam)
            {
                SendMessage(GetParent(hDlg), PSM_SETCURSEL, (s_iCurrentTab + 1) % (g_fFilterPresent ? _countof(rgpTab) : _countof(rgpTab) - 1), NULL);
            }
            else if (IDB_PREV_TAB == wParam)
            {
                SendMessage(GetParent(hDlg),
                    PSM_SETCURSEL,
                    (s_iCurrentTab ? s_iCurrentTab - 1 : (g_fFilterPresent ? _countof(rgpTab) - 1 : _countof(rgpTab) - 2)),
                    NULL);
            }
            else if (BN_CLICKED == HIWORD(wParam))
            {
                switch (LOWORD(wParam))
                {
                    case IDC_GESTURE_DM:
                        // If checking DM, clear PAN and SCROLL
                        if (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_GESTURE_DM, BM_GETCHECK, 0, 0))
                        {
                            SendDlgItemMessage(hDlg, IDC_GESTURE_PAN, BM_SETCHECK, 0, 0);
                            SendDlgItemMessage(hDlg, IDC_GESTURE_SCROLL, BM_SETCHECK, 0, 0);
                        }
                        break;
                    case IDC_GESTURE_PAN:
                    case IDC_GESTURE_SCROLL:
                        // If checking PAN or SCROLL, clear DM
                        if (BST_CHECKED == SendDlgItemMessage(hDlg, LOWORD(wParam), BM_GETCHECK, 0, 0))
                        {
                            SendDlgItemMessage(hDlg, IDC_GESTURE_DM, BM_SETCHECK, 0, 0);
                        }
                        break;
                }
            }
            break;
        case WM_NOTIFY:
            switch (((LPNMHDR)lParam)->code)
            {
                case PSN_SETACTIVE:
                    s_iCurrentTab = 3;
                    break;
                case PSN_APPLY:
                    pSettings->dwShowDM = (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_GESTURE_DM, BM_GETCHECK, 0, 0));
                    pSettings->dwShowPAN = (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_GESTURE_PAN, BM_GETCHECK, 0, 0));
                    pSettings->dwShowSCROLL = (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_GESTURE_SCROLL, BM_GETCHECK, 0, 0));
                    pSettings->dwShowSELECT = (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_GESTURE_SELECT, BM_GETCHECK, 0, 0));
                    pSettings->dwShowDSELECT = (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_GESTURE_DSELECT, BM_GETCHECK, 0, 0));
                    pSettings->dwShowHOLD = (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_GESTURE_HOLD, BM_GETCHECK, 0, 0));

                    GetDlgItemText(hDlg, IDC_GESTURE_DM_COLOR, szTmp, _countof(szTmp));
                    if (!ValidateColor(szTmp, _countof(szTmp), &pSettings->dwDMColor))
                    {
                        MessageBox(hDlg, L"DIRECT MANIPULATION Color is invalid.  Please specify a valid color (BBGGRR).", L"Invalid Color", MB_OK);
                        SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID);
                        return TRUE;
                    }
                    GetDlgItemText(hDlg, IDC_GESTURE_PAN_COLOR, szTmp, _countof(szTmp));
                    if (!ValidateColor(szTmp, _countof(szTmp), &pSettings->dwPANColor))
                    {
                        MessageBox(hDlg, L"PAN Color is invalid.  Please specify a valid color (BBGGRR).", L"Invalid Color", MB_OK);
                        SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID);
                        return TRUE;
                    }
                    GetDlgItemText(hDlg, IDC_GESTURE_SCROLL_COLOR, szTmp, _countof(szTmp));
                    if (!ValidateColor(szTmp, _countof(szTmp), &pSettings->dwSCROLLColor))
                    {
                        MessageBox(hDlg, L"SCROLL Color is invalid.  Please specify a valid color (BBGGRR).", L"Invalid Color", MB_OK);
                        SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID);
                        return TRUE;
                    }
                    GetDlgItemText(hDlg, IDC_GESTURE_SELECT_COLOR, szTmp, _countof(szTmp));
                    if (!ValidateColor(szTmp, _countof(szTmp), &pSettings->dwSELECTColor))
                    {
                        MessageBox(hDlg, L"SELECT Color is invalid.  Please specify a valid color (BBGGRR).", L"Invalid Color", MB_OK);
                        SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID);
                        return TRUE;
                    }
                    GetDlgItemText(hDlg, IDC_GESTURE_DSELECT_COLOR, szTmp, _countof(szTmp));
                    if (!ValidateColor(szTmp, _countof(szTmp), &pSettings->dwDSELECTColor))
                    {
                        MessageBox(hDlg, L"DOUBLE SELECT Color is invalid.  Please specify a valid color (BBGGRR).", L"Invalid Color", MB_OK);
                        SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID);
                        return TRUE;
                    }
                    GetDlgItemText(hDlg, IDC_GESTURE_HOLD_COLOR, szTmp, _countof(szTmp));
                    if (!ValidateColor(szTmp, _countof(szTmp), &pSettings->dwHOLDColor))
                    {
                        MessageBox(hDlg, L"HOLD Color is invalid.  Please specify a valid color (BBGGRR).", L"Invalid Color", MB_OK);
                        SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID);
                        return TRUE;
                    }

                    // Insure we have proper gesture messages enabled
                    if (pSettings->dwShowDM)
                    {
                        ASSERT(0 == pSettings->dwShowPAN);
                        ASSERT(0 == pSettings->dwShowSCROLL);

                        SwapDM(g_hwnd, TRUE);
                    }
                    else
                    {
                        SwapDM(g_hwnd, FALSE);
                    }
                    break;
            }
            break;
    }

    return FALSE;
}

// DlgProc for Warnings tab
BOOL WarningsTabDlgProc ( HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    PROPSHEETPAGE *pSheet = NULL;
    static REGSETTINGS *pSettings = NULL;
    WCHAR szTmp[MAX_PATH];

    switch (uMsg)
    {
        case WM_INITDIALOG:
            ASSERT(NULL != lParam);
            pSheet = (PROPSHEETPAGE *)lParam;
            ASSERT(NULL != pSheet->lParam);
            pSettings = (REGSETTINGS *)pSheet->lParam;

            SendDlgItemMessage(hDlg, IDC_WARNINGS_INITIAL, BM_SETCHECK, pSettings->dwShowInitialWarning, 0);
            SendDlgItemMessage(hDlg, IDC_WARNINGS_ID, BM_SETCHECK, pSettings->dwShowIDWarning, 0);
            SendDlgItemMessage(hDlg, IDC_WARNINGS_CONTACTS, BM_SETCHECK, pSettings->dwShowContactsWarning, 0);
            StringCchPrintf(szTmp, _countof(szTmp), L"%d", pSettings->dwContactsMin);
            SetDlgItemText(hDlg, IDC_WARNINGS_CONTACTS_MIN, szTmp);
            StringCchPrintf(szTmp, _countof(szTmp), L"%d", pSettings->dwContactsMax);
            SetDlgItemText(hDlg, IDC_WARNINGS_CONTACTS_MAX, szTmp);

            if (!g_fFilterPresent)
            {
                EnableWindow(GetDlgItem(hDlg, IDC_WARNINGS_ID), FALSE);
                EnableWindow(GetDlgItem(hDlg, IDC_WARNINGS_CONTACTS), FALSE);
            }

            if (!g_fFilterPresent || (0 == pSettings->dwShowContactsWarning))
            {
                EnableWindow(GetDlgItem(hDlg, IDC_WARNINGS_CONTACTS_MIN), FALSE);
                EnableWindow(GetDlgItem(hDlg, IDC_WARNINGS_CONTACTS_MAX), FALSE);
                EnableWindow(GetDlgItem(hDlg, IDC_WARNINGS_TEXT1), FALSE);
                EnableWindow(GetDlgItem(hDlg, IDC_WARNINGS_TEXT2), FALSE);
                EnableWindow(GetDlgItem(hDlg, IDC_WARNINGS_TEXT3), FALSE);
            }
            break;
        case WM_DESTROY:
            break;
        case WM_COMMAND:
            if (IDB_EXIT_APP == wParam)
            {
                if (IDYES == MessageBox(g_hwnd, L"Are you sure you want to exit CETouchView?", L"Exit Application", MB_YESNO))
                {
                    DestroyWindow(g_hwnd);
                }
            }
            else if (IDB_OK == wParam)
            {
                // Tell the property sheet to save changes
                PostMessage(GetParent(hDlg), WM_COMMAND, IDOK, NULL);
            }
            else if (IDB_CANCEL == wParam)
            {
                // Tell the property sheet to cancel changes
                PostMessage(GetParent(hDlg), WM_COMMAND, IDCANCEL, NULL);
            }
            else if (IDB_NEXT_TAB == wParam)
            {
                SendMessage(GetParent(hDlg), PSM_SETCURSEL, (s_iCurrentTab + 1) % (g_fFilterPresent ? _countof(rgpTab) : _countof(rgpTab) - 1), NULL);
            }
            else if (IDB_PREV_TAB == wParam)
            {
                SendMessage(GetParent(hDlg),
                    PSM_SETCURSEL,
                    (s_iCurrentTab ? s_iCurrentTab - 1 : (g_fFilterPresent ? _countof(rgpTab) - 1 : _countof(rgpTab) - 2)),
                    NULL);
            }
            else if ((BN_CLICKED == HIWORD(wParam)) && (IDC_WARNINGS_CONTACTS == LOWORD(wParam)))
            {
                // If checking checkbox, enable edit controls, else disable them
                BOOL fEnable = (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_WARNINGS_CONTACTS, BM_GETCHECK, 0, 0));
                EnableWindow(GetDlgItem(hDlg, IDC_WARNINGS_CONTACTS_MIN), fEnable);
                EnableWindow(GetDlgItem(hDlg, IDC_WARNINGS_CONTACTS_MAX), fEnable);
                EnableWindow(GetDlgItem(hDlg, IDC_WARNINGS_TEXT1), fEnable);
                EnableWindow(GetDlgItem(hDlg, IDC_WARNINGS_TEXT2), fEnable);
                EnableWindow(GetDlgItem(hDlg, IDC_WARNINGS_TEXT3), fEnable);
            }
            break;
        case WM_NOTIFY:
            switch (((LPNMHDR)lParam)->code)
            {
                case PSN_SETACTIVE:
                    s_iCurrentTab = 4;
                    break;
                case PSN_APPLY:
                    pSettings->dwShowInitialWarning = (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_WARNINGS_INITIAL, BM_GETCHECK, 0, 0));
                    pSettings->dwShowIDWarning = (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_WARNINGS_ID, BM_GETCHECK, 0, 0));
                    pSettings->dwShowContactsWarning = (BST_CHECKED == SendDlgItemMessage(hDlg, IDC_WARNINGS_CONTACTS, BM_GETCHECK, 0, 0));

                    GetDlgItemText(hDlg, IDC_WARNINGS_CONTACTS_MIN, szTmp, _countof(szTmp));
                    DWORD dwMin, dwMax;
                    BOOL fMin = ValidateNumber(szTmp, _countof(szTmp), 0, max(g_dwScreeW,g_dwScreeH) * 4, &dwMin);
                    GetDlgItemText(hDlg, IDC_WARNINGS_CONTACTS_MAX, szTmp, _countof(szTmp));
                    BOOL fMax = ValidateNumber(szTmp, _countof(szTmp), 0, max(g_dwScreeW,g_dwScreeH) * 4, &dwMax);
                    if (!fMin || !fMax || (dwMin > dwMax))
                    {
                        StringCchPrintf(szTmp, _countof(szTmp), L"This needs to be true:\n\n0 <= MIN <= MAX <= %d", max(g_dwScreeW,g_dwScreeH) * 4);
                        MessageBox(hDlg, szTmp, L"Invalid Min/Max Values", MB_OK);
                        SetWindowLong(hDlg, DWL_MSGRESULT, PSNRET_INVALID);
                        return TRUE;
                    }
                    pSettings->dwContactsMin = dwMin;
                    pSettings->dwContactsMax = dwMax;
                    break;
            }
            break;
    }

    return FALSE;
}

// DlgProc for Record/Playback tab
BOOL RecPlayTabDlgProc ( HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    PROPSHEETPAGE *pSheet = NULL;
    static REGSETTINGS *pSettings = NULL;
    WCHAR szTmp[MAX_PATH];

    switch (uMsg)
    {
        case WM_INITDIALOG:
            ASSERT(NULL != lParam);
            pSheet = (PROPSHEETPAGE *)lParam;
            ASSERT(NULL != pSheet->lParam);
            pSettings = (REGSETTINGS *)pSheet->lParam;

            SetDlgItemText(hDlg, IDC_RECPLAY_FILENAME, pSettings->szFileName);
            break;
        case WM_DESTROY:
            break;
        case WM_COMMAND:
            if ((EN_UPDATE == HIWORD(wParam)) && (IDC_RECPLAY_FILENAME == LOWORD(wParam)))
            {
                GetDlgItemText(hDlg, IDC_RECPLAY_FILENAME, szTmp, _countof(szTmp));
                if (*szTmp)
                {
                    SetDlgItemText(hDlg, IDB_RECPLAY_PLAYBACK, L"Begin Playback");
                    EnableWindow(GetDlgItem(hDlg, IDB_RECPLAY_RECORD), TRUE);
                }
                else
                {
                    SetDlgItemText(hDlg, IDB_RECPLAY_PLAYBACK, L"Begin Puppet");
                    EnableWindow(GetDlgItem(hDlg, IDB_RECPLAY_RECORD), FALSE);
                }
            }
            else switch (wParam)
            {
                case IDB_EXIT_APP:
                    if (IDYES == MessageBox(g_hwnd, L"Are you sure you want to exit CETouchView?", L"Exit Application", MB_YESNO))
                    {
                        DestroyWindow(g_hwnd);
                    }
                    break;
                case IDB_OK:
                    // Tell the property sheet to save changes
                    PostMessage(GetParent(hDlg), WM_COMMAND, IDOK, NULL);
                    break;
                case IDB_CANCEL:
                    // Tell the property sheet to cancel changes
                    PostMessage(GetParent(hDlg), WM_COMMAND, IDCANCEL, NULL);
                    break;
                case IDB_NEXT_TAB:
                    SendMessage(GetParent(hDlg), PSM_SETCURSEL, (s_iCurrentTab + 1) % (g_fFilterPresent ? _countof(rgpTab) : _countof(rgpTab) - 1), NULL);
                    break;
                case IDB_PREV_TAB:
                    SendMessage(GetParent(hDlg),
                        PSM_SETCURSEL,
                        (s_iCurrentTab ? s_iCurrentTab - 1 : (g_fFilterPresent ? _countof(rgpTab) - 1 : _countof(rgpTab) - 2)),
                        NULL);
                    break;
                case IDB_RECPLAY_PLAYBACK:
                    GetDlgItemText(hDlg, IDC_RECPLAY_FILENAME, szTmp, _countof(szTmp));

                    // this message will come first before the update notification, which
                    // will handle the registry update, right now let's just validate that
                    // the filename is empty, or if filled in, that it exists

                    if (L'\0' != *szTmp)
                    {
                        FILE *f = NULL;
                        if (0 != _wfopen_s(&f, szTmp, L"rt"))
                        {
                            MessageBox(hDlg, L"Unable to open the file specified.  Nothing to play back.", L"Invalid Filename", MB_OK);
                            return FALSE;
                        }
                        fclose(f);
                    }

                    pSettings->dwMacroMode = 0;     // set for playback

                    // Tell the property sheet to save changes
                    PostMessage(GetParent(hDlg), WM_COMMAND, IDOK, NULL);
                    // Tell the main window to signal the Filter
                    PostMessage(GetParent(GetParent(hDlg)), WM_TRIGGER_FILTER, NULL, NULL);
                    break;
                case IDB_RECPLAY_RECORD:
                    GetDlgItemText(hDlg, IDC_RECPLAY_FILENAME, szTmp, _countof(szTmp));

                    if (L'\0' != *szTmp)
                    {
                        FILE *f = NULL;
                        if (0 != _wfopen_s(&f, szTmp, L"wt"))
                        {
                            MessageBox(hDlg, L"Unable to create the file specified.", L"Invalid Filename", MB_OK);
                            return FALSE;
                        }
                        fclose(f);
                    }
                    else
                    {
                        MessageBox(hDlg, L"Specify a valid file name!", L"Invalid Filename", MB_OK);
                        return FALSE;
                    }

                    g_fRecording = pSettings->dwMacroMode = 1;

                    // Tell the property sheet to save changes
                    PostMessage(GetParent(hDlg), WM_COMMAND, IDOK, NULL);
                    // Tell the main window to signal the Filter
                    PostMessage(GetParent(GetParent(hDlg)), WM_TRIGGER_FILTER, NULL, NULL);
                    break;
            }
            break;
        case WM_NOTIFY:
            switch (((LPNMHDR)lParam)->code)
            {
                case PSN_SETACTIVE:
                    s_iCurrentTab = 5;
                    break;
                case PSN_APPLY:
                    GetDlgItemText(hDlg, IDC_RECPLAY_FILENAME, szTmp, _countof(szTmp));
                    if (wcscmp(szTmp, pSettings->szFileName))
                    {
                        LocalFree((HLOCAL)pSettings->szFileName);
                        pSettings->szFileName = StrDup(szTmp);
                    }
                    break;
            }
            break;
    }

    return FALSE;
}
